import re
import io
import csv
import datetime
import time
import logging

from congress.tasks import utils
from congress.tasks.vote_info import output_vote

# load some hard-coded codes
special_vote_options = { }
for rec in csv.reader(open("tasks/voteview_codedoptions.csv")):
    if rec[0] == "vote date": continue # header
    special_vote_options[rec[1]] = (rec[2], dict((int(r.split(':', 1)[0]), r.split(':', 1)[1]) for r in rec[3].split(';')))


def run(options):
    congress = options.get("congress", None)
    congress = int(congress) if congress else utils.current_congress()

    chamber = options.get('chamber', None)

    # we're going to need to map votes to sessions because in modern history the numbering resets by session
    session_dates = list(csv.DictReader(io.StringIO(utils.download("http://www.govtrack.us/data/us/sessions.tsv").encode("utf8")), delimiter="\t"))

    # download the vote data now
    if chamber and chamber in [ "h", "s" ]:
        votes = get_votes(chamber, congress, options, session_dates)
    else:
        votes = get_votes("h", congress, options, session_dates) + get_votes("s", congress, options, session_dates)

    utils.process_set(votes, put_vote, options)


def vote_list_source_urls_for(congress, chamber, options):
    url = "http://www.voteview.com/%s%02d.htm" % (("house" if chamber == "h" else "senate"), congress)
    index_page = utils.download(url, cache_file_for(congress, chamber, "html"), options)
    if index_page == None:
        raise Exception("No data.")  # should only happen on a 404

    def match(pattern):
        matches = re.findall(pattern, index_page, re.I)
        if len(matches) != 1:
            raise ValueError("Index page %s did not match one value for pattern %s." % (url, pattern))
        return matches[0]

    return match("ftp://voteview.com/[^\.\s]+\.ord"), match("ftp://voteview.com/dtl/[^\.\s]+\.dtl")


def cache_file_for(congress, chamber, file_type):
    return "voteview/%s-%s.%s" % (congress, chamber, file_type)


def get_state_from_icpsr_state_code(icpsr_state_code):
    icpsr_state_code_map = {
        1: "CT",
        2: "ME",
        3: "MA",
        4: "NH",
        5: "RI",
        6: "VT",
        11: "DE",
        12: "NJ",
        13: "NY",
        14: "PA",
        21: "IL",
        22: "IN",
        23: "MI",
        24: "OH",
        25: "WI",
        31: "IA",
        32: "KS",
        33: "MN",
        34: "MO",
        35: "NE",
        36: "ND",
        37: "SD",
        40: "VA",
        41: "AL",
        42: "AR",
        43: "FL",
        44: "GA",
        45: "LA",
        46: "MS",
        47: "NC",
        48: "SC",
        49: "TX",
        51: "KY",
        52: "MD",
        53: "OK",
        54: "TN",
        55: "DC",
        56: "WV",
        61: "AZ",
        62: "CO",
        63: "ID",
        64: "MT",
        65: "NV",
        66: "NM",
        67: "UT",
        68: "WY",
        71: "CA",
        72: "OR",
        73: "WA",
        81: "AK",
        82: "HI",
        99: None,  # Used by presidents
    }

    return icpsr_state_code_map[icpsr_state_code]


def get_party_from_icpsr_party_code(icpsr_party_code):
    icpsr_party_code_map = {
        1: "Federalist",
        9: "Jefferson Republican",
        10: "Anti-Federalist",
        11: "Jefferson Democrat",
        13: "Democrat-Republican",
        22: "Adams",
        25: "National Republican",
        26: "Anti Masonic",
        29: "Whig",
        34: "Whig and Democrat",
        37: "Constitutional Unionist",
        40: "Anti-Democrat and States Rights",
        41: "Anti-Jackson Democrat",
        43: "Calhoun Nullifier",
        44: "Nullifier",
        46: "States Rights",
        48: "States Rights Whig",
        100: "Democrat",
        101: "Jackson Democrat",
        103: "Democrat and Anti-Mason",
        104: "Van Buren Democrat",
        105: "Conservative Democrat",
        108: "Anti-Lecompton Democrat",
        110: "Popular Sovereignty Democrat",
        112: "Conservative",
        114: "Readjuster",
        117: "Readjuster Democrat",
        118: "Tariff for Revenue Democrat",
        119: "United Democrat",
        200: "Republican",
        202: "Union Conservative",
        203: "Unconditional Unionist",
        206: "Unionist",
        208: "Liberal Republican",
        212: "United Republican",
        213: "Progressive Republican",
        214: "Non-Partisan and Republican",
        215: "War Democrat",
        300: "Free Soil",
        301: "Free Soil Democrat",
        302: "Free Soil Whig",
        304: "Anti-Slavery",
        308: "Free Soil American and Democrat",
        310: "American",
        326: "National Greenbacker",
        328: "Independent",
        329: "Ind. Democrat",
        331: "Ind. Republican",
        333: "Ind. Republican-Democrat",
        336: "Anti-Monopolist",
        337: "Anti-Monopoly Democrat",
        340: "Populist",
        341: "People's",
        347: "Prohibitionist",
        353: "Ind. Silver Republican",
        354: "Silver Republican",
        355: "Union",
        356: "Union Labor",
        370: "Progressive",
        380: "Socialist",
        401: "Fusionist",
        402: "Liberal",
        403: "Law and Order",
        522: "American Labor",
        537: "Farmer-Labor",
        555: "Jackson",
        603: "Ind. Whig",
        1060: "Silver",
        1061: "Emancipationist",
        1111: "Liberty",
        1116: "Conservative Republican",
        1275: "Anti-Jackson",
        1346: "Jackson Republican",
        3333: "Opposition",
        4000: "Anti-Administration",
        4444: "Union",
        5000: "Pro-Administration",
        6000: "Crawford Federalist",
        6666: "Crawford Republican",
        7000: "Jackson Federalist",
        7777: "Crawford Republican",
        8000: "Adams-Clay Federalist",
        8888: "Adams-Clay Republican",
        9000: "Unknown",
        9999: "Unknown",
    }

    return icpsr_party_code_map.get(icpsr_party_code)


def parse_voteview_vote_code(vote_code):
    # Convert the integer codes into a tuple containing:
    #    standard vote options "Yea", "Nay", "Not Voting", "Present"
    #    an additional string so that we don't lose any information provided by voteview
    # Probably the House used Aye and No in some votes, but we don't
    # know which. "Yea" and "Nay" are always used by the Senate, and always
    # in the House on the passage of bills.
    # A paired vote is when two members plan to be absent in a way that
    # does not affect the vote outcome. You can see in the Congressional
    # record who is paired with who. Sometimes the pairings are for a
    # particular vote, other pairings are "until further notice." The paired
    # members are recorded as not voting. A live pair is when one half of
    # the pair is present and withdraws their vote and votes present because
    # the other half of the pair isn't there. Live pairs aren't recorded
    # in this data and are treated simply as pairs (and thus for us, not
    # voting). Some paired members are recorded simply as present in this
    # data --- not clear why that would be.
    # See the House vote on the Civil Rights Act of 1957 (85th Congress,
    # Jun 18, 1957, what this data calls #42, volume 103 page 9518 of
    # the Congressional Record) for an example of paired votes.
    vote_code_map = {
        0: (None, None),  # not a member at the time of the vote (but sometimes recorded as Not Voting)
        1: ("Yea", None),
        2: ("Not Voting", "paired-yea"),
        3: ("Not Voting", "announced-yea"),
        4: ("Not Voting", "announced-nay"),
        5: ("Not Voting", "paired-nay"),
        6: ("Nay", None),
        7: ("Present", "type-seven"),
        8: ("Present", "type-eight"),
        9: ("Not Voting", None),
    }
    return vote_code_map[vote_code]


def parse_vote_list_line(vote_list_line):
    return re.match(r"^([\s\d]{2}\d)([\s\d]{4}\d)([\s\d]\d)([\s\d]{2})([^\d]+?)([\s\d]{3}\d)([\s\d])([\s\d])([^\s\d][^\d]+?(?:\d\s+)?)(\d+)$", vote_list_line).groups()


def parse_rollcall_dtl_list_line(rollcall_list_line):
    return re.match(r"^([\s\d]{3}\d)([\s\d]{4}\d)?([\s\d]\d)\s(.*?)\s*$", rollcall_list_line).groups()


def parse_rollcall_dtl_list_first_line(rollcall_dtl_first_line):
    return re.match(r"^(.{14})(.{15})(.{10})?(.+?)(?:\s{3,}\d{2,3})?$", rollcall_dtl_first_line).groups()


def parse_rollcall_dtl_date(rollcall_dtl_date):
    from datetime import datetime

    potential_date_formats = [
        "%b %d, %Y",  # JAN 1, 1900
        "%B %d, %Y",  # JANUARY 1, 1900
        "%b, %d, %Y",  # JAN, 1, 1900
        "%B, %d, %Y",  # JANUARY, 1, 1900
        "%b.%d, %Y",  # JAN.1, 1900
    ]

    # Make things easier by removing periods after month abbreviations.
    rollcall_dtl_date = rollcall_dtl_date.replace(". ", " ")

    # Make things easier by inserting spaces after commas where they are missing.
    rollcall_dtl_date = rollcall_dtl_date.replace(",1", ", 1")

    # Python doesn't consider "SEPT" a valid abbreviation for September.
    rollcall_dtl_date = rollcall_dtl_date.replace("SEPT ", "SEP ")

    parsed_date = None

    for potential_date_format in potential_date_formats:
        try:
            parsed_date = datetime.strptime(rollcall_dtl_date, potential_date_format)
        except ValueError:
            pass
        else:
            break

    formatted_date = utils.format_datetime(parsed_date)

    return formatted_date[:10] if formatted_date is not None else formatted_date


def extract_vote_info_from_parsed_vote_list_line(parsed_vote_list_line):
    vote_info = {
        "congress": int(parsed_vote_list_line[0]) if parsed_vote_list_line[0].strip() else None,
        "icpsr_id": int(parsed_vote_list_line[1]) if parsed_vote_list_line[1].strip() else None,
        "icpsr_state": int(parsed_vote_list_line[2]) if parsed_vote_list_line[2].strip() else None,
        "district": int(parsed_vote_list_line[3]) if parsed_vote_list_line[3].strip() else None,
        # parsed_vote_list_line[4] is partial state name
        "state_name": parsed_vote_list_line[4].strip(),
        "icpsr_party": int(parsed_vote_list_line[5]) if parsed_vote_list_line[5].strip() else None,
        "occupancy": int(parsed_vote_list_line[6]) if parsed_vote_list_line[6].strip() else None,
        "means": int(parsed_vote_list_line[7]) if parsed_vote_list_line[7].strip() else None,
        # parsed_vote_list_line[8] is partial member name
        "member_name": parsed_vote_list_line[8].strip(),
        "votes": [int(icpsr_vote_code) for icpsr_vote_code in parsed_vote_list_line[9]],
    }

    return vote_info


def extract_rollcall_info_from_parsed_rollcall_dtl_list_line(parsed_rollcall_dtl_list_line):
    rollcall_info = {
        "vote": int(parsed_rollcall_dtl_list_line[0]),
        "line": int(parsed_rollcall_dtl_list_line[2]),
        "text": parsed_rollcall_dtl_list_line[3],
    }

    return rollcall_info


def parse_vote_list_file(vote_list_file):
    # Each line in the vote list file is for a Member of Congress, with
    # identifying data in the left column followed by one character per
    # vote (1=aye, etc.).

    logging.info("Parsing vote list file...")

    vote_list_info = []

    for vote_list_line in vote_list_file.split("\r\n"):
        if not vote_list_line.strip():
            continue

        vote_info = extract_vote_info_from_parsed_vote_list_line(parse_vote_list_line(vote_list_line))

        vote_info["state"] = get_state_from_icpsr_state_code(vote_info["icpsr_state"]) if vote_info["icpsr_state"] is not None else None
        vote_info["party"] = get_party_from_icpsr_party_code(vote_info["icpsr_party"]) if vote_info["icpsr_party"] is not None else None

        icpsr_id = vote_info["icpsr_id"]

        # I think these are mistakes? Don't know if the 9- codes something special.
        if icpsr_id == 91449: icpsr_id = 1449
        if icpsr_id == 92484: icpsr_id = 2484
        if icpsr_id == 94804: icpsr_id = 4804
        if icpsr_id == 94891: icpsr_id = 4891
        if icpsr_id == 96738: icpsr_id = 6738
        if icpsr_id == 98500: icpsr_id = 8500
        if icpsr_id == 99369: icpsr_id = 9369
        if icpsr_id == 90618: icpsr_id = 10618
        if icpsr_id == 90634: icpsr_id = 10634
        if icpsr_id == 91043: icpsr_id = 11043
        if icpsr_id == 93033: icpsr_id = 13033
        if icpsr_id == 94428: icpsr_id = 14428
        if icpsr_id == 94454: icpsr_id = 14454
        if icpsr_id == 94602: icpsr_id = 14602
        if icpsr_id == 94628: icpsr_id = 14628
        if icpsr_id == 95122: icpsr_id = 15122
        if icpsr_id == 95415: icpsr_id = 15415
        if icpsr_id == 3769: icpsr_id = 15101 # guy was given two ids
        if icpsr_id == 14240: icpsr_id = 94240 # per our id

        try:
            bioguide_id = utils.get_person_id("icpsr" if vote_info["state_name"] != "USA" else "icpsr_prez", icpsr_id, "bioguide")
        except KeyError as e:
            # skip some guys named Poe (99999) and Chambers (10509) that don't seem to have existed and didn't cast actual votes,
            # and Jack Swigert (15067) who died before being sworn in.
            # and presidents may not have bioguide IDs
            if icpsr_id not in (99999, 10509, 15067) and vote_info["state_name"] != "USA":
                logging.error("Problem with member %s ([%d] %s) of %s %s: %s" % (vote_info["member_name"], vote_info["icpsr_party"], vote_info["party"],
                                                                             vote_info["state_name"], vote_info["district"], e.message))
                #logging.error(vote_info)
            bioguide_id = None
        else:
            logging.debug("Parsed member %s ([%d] %s) of %s %s..." % (vote_info["member_name"], vote_info["icpsr_party"], vote_info["party"],
                                                                      vote_info["state_name"], vote_info["district"]))
        vote_info["bioguide_id"] = bioguide_id

        # This is used to record the President's position, or something.
        # Mark this record so build_votes can separated it out from Member votes.
        vote_info["is_president"] = True if vote_info["icpsr_state"] == 99 else False

        vote_list_info.append(vote_info)

    return vote_list_info


def parse_rollcall_dtl_list_file(rollcall_dtl_list_file, congress):
    rollcall_dtl_list_info = {}

    for rollcall_dtl_list_line in rollcall_dtl_list_file.split("\r\n"):
        if not rollcall_dtl_list_line.strip():
            continue

        rollcall_dtl_list_line_info = extract_rollcall_info_from_parsed_rollcall_dtl_list_line(parse_rollcall_dtl_list_line(rollcall_dtl_list_line))

        if rollcall_dtl_list_line_info["line"] == 1:
            rollcall_info = {}

            rollcall_dtl_list_first_line_parts = parse_rollcall_dtl_list_first_line(rollcall_dtl_list_line_info["text"])
            rollcall_info["record_id"] = rollcall_dtl_list_first_line_parts[0].strip()
            rollcall_info["journal_id"] = rollcall_dtl_list_first_line_parts[1].strip()

            rollcall_info["date_unparsed"] = rollcall_dtl_list_first_line_parts[3].strip()
            rollcall_info["date"] = parse_rollcall_dtl_date(rollcall_info["date_unparsed"])

            rollcall_info["bill_unparsed"] = rollcall_dtl_list_first_line_parts[2].strip()
            m = re.match(r"([A-Z]+)([0-9]+)$", rollcall_info["bill_unparsed"])
            if m:
                bill_type_map = {
                   'HR': 'hr', 'H': 'hr',
                   'S': 's',
                   'HJR': 'hjres', 'HJ': 'hjres', 'HJRE': 'hjres', 'HJRES': 'hjres',
                   'SJR': 'sjres', 'SJ': 'sjres', 'SJRE': 'sjres', 'SJRES': 'sjres',
                   'HCR': 'hconres', 'HCRE': 'hconres', 'HCRES': 'hconres', 'HCONR': 'hconres', 'HCON': 'hconres',
                   'SCR': 'sconres', 'SCRE': 'sconres', 'SCRES': 'sconres', 'SCONRES': 'sconres', 'SCONR': 'sconres', 'SCON': 'sconres',
                   'HRE': 'hres', 'HRES': 'hres',
                   'SRE': 'sres', 'SR': 'sres', 'SRES': 'sres' }
                if not m.group(1) in bill_type_map:
                    logging.error('Could not parse bill: %s' % rollcall_info["bill_unparsed"])
                else:
                    rollcall_info["bill"] = { 'congress': congress, 'type': bill_type_map[m.group(1)], 'number': int(m.group(2)) }

        elif rollcall_dtl_list_line_info["line"] == 2:
            pass
        elif rollcall_dtl_list_line_info["line"] == 3:
            rollcall_info["description"] = rollcall_dtl_list_line_info["text"]
        else:
            rollcall_info["description"] += " " + rollcall_dtl_list_line_info["text"]

        rollcall_dtl_list_info[rollcall_dtl_list_line_info["vote"]] = rollcall_info

    return rollcall_dtl_list_info


def build_votes(vote_list):
    # Go from a list of individuals (and their votes) to a mapping
    # from votes to how the individuals voted on it.

    logging.info("Building votes...")

    votes = {}
    presidents_positions = {}

    for voter in vote_list:
        for i, choice in enumerate(voter["votes"]):
            # Separate the president's position from Member votes.
            if voter["is_president"]:
                presidents_positions[i] = choice
                continue

            # Drop anyone we didn't have a bioguide id for. We issued warnings
            # when we did the lookup if we couldn't find the id. Any remaining
            # cases are individuals who didn't actually take office and didn't
            # actually vote. Presidents may not have bioguide IDs so we filter
            # those first above.
            if voter["bioguide_id"] is None:
                continue

            # Make a record for this vote, grouped by vote option (Aye, etc).
            votes.setdefault(i, []).append({
                "id": voter["bioguide_id"],
                "display_name": voter["member_name"],
                "party": voter["party"],
                "state": voter["state"],
                "vote": choice,
            })

    # sort for output
    for voters in votes.values():
        voters.sort(key=lambda v: v['display_name'])

    return (votes, presidents_positions)


def session_from_date(date, session_dates):
    for sess in session_dates:
        if sess["start"] <= date <= sess["end"]:
            return int(sess["congress"]), sess["session"]
    return None, None

def parse_rollcall_description(rollcall):
    # The description sometimes has additional metadata. It's a little tricky
    # to parse because the description has hyphens at the ends of lines where
    # words are split.
    dparts = rollcall['description'].split(". ")
    while len(dparts) > 1:
        dpart = dparts[-1].strip(".- ") # remove trailing spaces, hyphens, and periods (which occur at the end of the final dpart but not inner ones because it is the split string)
        if dpart == "NAY SUPPORTS PRESIDENT'S POSITION":
            rollcall['presidents_position'] =  { "option": "Nay" } # also recorded in the big table, so we probably already have this
        elif dpart == "YEA SUPPORTS PRESIDENT'S POSITION":
            rollcall['presidents_position'] =  { "option": "Yea" }
        elif dpart in ("REJECTED", "PASSED", "AGREED TO", "ADOPTED", "ACCEPTED", "CONFIRMED", "RATIFIED"):
            rollcall['result'] = dpart.title()
        elif dpart.startswith("(SEE CQ "):
            pass # remove this
        else:
            # Unrecognized, so stop here.
            break
        # Remove this part from the description.
        dparts.pop(-1)
    rollcall['description'] = ". ".join(dparts)
    if not rollcall['description'].endswith('.'): rollcall['description'] += "."

def build_votes_dict(votes_list, rollcall):    
    if rollcall.get("description") in special_vote_options:
        # Some votes are for things besides aye/no etc where the vote
        # description says how the numeric codes are mapped to options.
        # e.g. for Election of the Speaker, 1 will be one candidate, 2
        # will be another candidate. We've manually coded these and
        # loaded them at the top of the module. In these cases, we also
        # have replacement strings for the vote description.
        original_description = rollcall["description"]
        new_description, vote_codes = special_vote_options[original_description]
        rollcall["description"] = new_description
        for v in votes_list:
            if v["vote"] == 0:
                v["vote"] = None
            elif v["vote"] == 9:
                v["vote"] = "Not Voting"
            else:
                try:
                    v["vote"] = vote_codes[v["vote"]]
                except KeyError:
                    logging.error('Vote "%s" had a "%d" vote.' % (original_description, v["vote"]))
                    v["vote"] = "Unknown"

    else:
        # This is a regular vote. Use the regular voteview codebook.
        for v in votes_list:
            v["vote"], v["voteview_votecode_extra"] = parse_voteview_vote_code(v["vote"])

    # Now make a dict from vote option to the legislators who voted
    # that option. Preserve ordering of votes_list which is already
    # sorted.
    ret = {
        choice: [v for v in votes_list if v["vote"] == choice]
        for choice in set(v["vote"] for v in votes_list)
        if choice != None # legislators who were not serving at the time of the vote
    }

    # No longer need the "vote" keys.
    for v in votes_list:
        del v["vote"]

    return ret

def get_votes(chamber, congress, options, session_dates):
    logging.warn("Getting votes for %d-%s..." % (congress, chamber))

    vote_list_url, rollcall_list_url = vote_list_source_urls_for(congress, chamber, options)

    # Load the ORD file which contains the matrix of how people voted.

    vote_list_file = utils.download(vote_list_url, cache_file_for(congress, chamber, "ord"), options).encode("utf-8")
    if not vote_list_file:
        logging.error("Couldn't download vote list file.")
        return None

    vote_list = parse_vote_list_file(vote_list_file)
    votes, presidents_positions = build_votes(vote_list)

    # Load the DTL file which lists each roll call vote with textual metadata.

    rollcall_list_file = utils.download(rollcall_list_url, cache_file_for(congress, chamber, "dtl"), options).encode("utf-8")
    if not rollcall_list_file:
        logging.error("Couldn't download rollcall list file.")
        return None
    rollcall_list = parse_rollcall_dtl_list_file(rollcall_list_file, congress)

    # Some dates are valid but incorrect. When the date doesn't even fall
    # within the Congress that we know the vote falls in, clear out the
    # date so we can try to guess a valid date in the next step.
    for rollcall_number in rollcall_list:
        rollcall = rollcall_list[rollcall_number]
        if rollcall["date"]:
            d_congress, d_session = session_from_date(rollcall["date"], session_dates)
            if d_congress != congress:
                rollcall["date"] = None

    # The dates listed in the DTL file were originally OCRd and have tons
    # of errors. Many strings could not be parsed. There are occasional
    # invalid dates (like Feb 29 on a non-leap year --- the 9s are probably
    # incorrectly OCR'd 5's). Try to resolve these quickly without resorting
    # to manual fact-checking...
    for i in range(min(rollcall_list)+1, max(rollcall_list) - 1):
        if rollcall_list[i]["date"]:
            continue  # was OK
        if not rollcall_list[i - 1]["date"]:
            continue  # preceding date not OK

        # If the vote is surrounded by votes on the same day, set the date to that day.
        if rollcall_list[i - 1]["date"] == rollcall_list[i + 1]["date"]:
            rollcall_list[i]["date"] = rollcall_list[i - 1]["date"]
            logging.error("Replacing %s with %s." % (rollcall_list[i]["date_unparsed"], rollcall_list[i - 1]["date"]))

        # Lump the vote with the previous date.
        else:
            rollcall_list[i]["date"] = rollcall_list[i - 1]["date"]
            logging.error("Replacing %s with %s (but might be as late as %s)." % (rollcall_list[i]["date_unparsed"], rollcall_list[i - 1]["date"], rollcall_list[i + 1]["date"]))

    # Form the output data.

    vote_output_list = []

    for rollcall_number in rollcall_list:
        vote_results = votes[rollcall_number - 1]
        rollcall = rollcall_list[rollcall_number]

        # Which session is this in? Compare the vote's date to the sessions.tsv file.
        if not rollcall["date"]:
            logging.error("Vote on %s was an invalid date, so we can't determine the session to save the file.. | %s" % (rollcall["date_unparsed"], rollcall["description"]))
            continue

        s_congress, session = session_from_date(rollcall["date"], session_dates)
        if s_congress != congress:
            # should not occur - handled above
            logging.error("Vote on %s disagrees about which Congress it is in." % rollcall["date"])
            continue
        if session is None:
            # This vote did not occur durring a session of Congress. Some sort of data error.
            logging.error("Vote on %s is not within a session of Congress." % rollcall["date"])
            continue

        # Only process votes from the requested session.
        if options.get("session") and session != options["session"]:
            continue

        rollcall['result'] = "unknown"
        if "description" in rollcall:
            parse_rollcall_description(rollcall)

        # Make the votes dictionary, but also replace the description
        # text when it contains coded vote information.
        votes_dict = build_votes_dict(vote_results, rollcall)

        # Form the vote dict.
        vote_output = {
            "vote_id": "%s%s-%d.%s" % (chamber, rollcall_number, congress, session),
            "source_url": "http://www.voteview.com",
            "updated_at": datetime.datetime.fromtimestamp(time.time()),

            "congress": congress,
            "session": session,
            "chamber": chamber,
            "number": rollcall_number,  # XXX: This is not the right number.
            "question": rollcall["description"] if "description" in rollcall else None,  # Sometimes there isn't a description.
            "type": normalize_vote_type(rollcall["description"]) if "description" in rollcall else None,
            "date": datetime.date(*[int(dd) for dd in rollcall["date"].split("-")]),  # turn YYYY-MM-DD into datetime.date() instance
            "date_unparsed": rollcall["date_unparsed"],
            "votes": votes_dict,
            "presidents_position": presidents_positions.get(rollcall_number) or rollcall.get('presidents_position'),
            "bill": rollcall.get('bill'),

            "category": "unknown",
            "requires": "unknown",
            "result": rollcall['result'],
        }

        vote_output_list.append(vote_output)

    return vote_output_list


def put_vote(vote, options):
    output_vote(vote, options, id_type="bioguide")
    return {"ok": True, "saved": True}


def normalize_vote_type(descr):
    if descr.startswith("TO PASS "):
        return "On Passage"
    if descr.startswith("TO AMEND "):
        return "On the Amendment"
    if descr.startswith("TO CONCUR IN THE SENATE AMENDMENT "):
        return "Concurring in the Senate Amendment"
    if descr.startswith("TO READ THE SECOND TIME "):
        return "Reading the Second Time"
    if descr.startswith("TO ADVISE AND CONSENT TO THE RATIFICATION OF THE TREATY"):
        return "On the Treaty"
    #logging.error("Unknown vote type: " + descr)
    return descr



# Convert GPO Fdsys STATUTE metadata into bill files.
#
# GPO has the Statutes at Large from 1951 (the 65th
# volume, 82nd Congress) to the present, with metadata
# at the level of the law.
#
# The bill files have sort of made up action entries
# since we don't know the legislative history of the bill.
# We also assume all bills are enacted by being signed
# by the President for the sake of outputting status
# information.
#
# First download the Statutes at Large from GPO:
#
# usc-run fdsys --collections=STATUTE --store=mods
#
# To process statute text, get the text PDFs:
#
# usc-run fdsys --collections=STATUTE --store=pdfs --granules
#
# Then run this script:
#
# usc-run statutes
#
# Processes all downloaded statutes files and saves bill files:
#   data/82/bills/hr/hr1/data.json and
#   data/82/bills/hr/hr1/text-versions/enr/data.json
#
# Specify --textversions to only write the text-versions file.
#
# If the individual statute PDF files are available, then
# additional options are possible:
#
# If --linkpdf is given, then *hard links* are created from
# where the PDF should be for bill text to where the PDF has
# been downloaded in the fdsys directory.
#
# If --extracttext is given, then the pdf is converted to text
# using "pdftotext -layout" and they are stored in files like
# data/82/bills/hr/hr1/text-versions/enr/document.txt. They are
# UTF-8 encoded and have form-feed characters marking page breaks.
#
# Examples:
# usc-run statutes --volume=65
# usc-run statutes --volumes=65-86
# usc-run statutes --year=1951
# usc-run statutes --years=1951-1972
# Processes just the indicated volume or range of volumes.
# Starting with the 93rd Congress (1973-1974, corresponding
# to volume 78 of the Statutes of Large), we have bill
# data from THOMAS. Be careful not to overwrite those files.
#
# With bill text missing from THOMAS/GPO from the 93rd to
# 102nd Congresses, fill in the text-versions files like so:
# usc-run statutes --volumes=87-106 --textversions

import logging
import time
import datetime
from lxml import etree
import glob
import json
import os.path
import subprocess

from congress.tasks import utils, bill_info, bill_versions
import fdsys


def run(options):
    root_dir = utils.data_dir() + '/fdsys/STATUTE'

    if "volume" in options:
        to_fetch = glob.glob(root_dir + "/*/STATUTE-" + str(int(options["volume"])))
    elif "volumes" in options:
        start, end = options["volumes"].split("-")
        to_fetch = []
        for v in range(int(start), int(end) + 1):
            to_fetch.extend(glob.glob(root_dir + "/*/STATUTE-" + str(v)))
    elif "year" in options:
        to_fetch = glob.glob(root_dir + "/" + str(int(options["year"])) + "/STATUTE-*")
    elif "years" in options:
        start, end = options["years"].split("-")
        to_fetch = []
        for y in range(int(start), int(end) + 1):
            to_fetch.extend(glob.glob(root_dir + "/" + str(y) + "/STATUTE-*"))
    else:
        to_fetch = sorted(glob.glob(root_dir + "/*/STATUTE-*"))

    logging.warn("Going to process %i volumes" % len(to_fetch))

    utils.process_set(to_fetch, proc_statute_volume, options)


def proc_statute_volume(path, options):
    mods = etree.parse(path + "/mods.xml")
    mods_ns = {"mods": "http://www.loc.gov/mods/v3"}

    # Load the THOMAS committee names for this Congress, which is our best
    # bet for normalizing committee names in the GPO data.
    congress = mods.find("/mods:extension[2]/mods:congress", mods_ns).text
    utils.fetch_committee_names(congress, options)

    logging.warn("Processing %s (Congress %s)" % (path, congress))

    package_id = mods.find("/mods:extension[2]/mods:accessId", mods_ns).text

    for bill in mods.findall("/mods:relatedItem", mods_ns):
        # MODS files also contain information about:
        # ['BACKMATTER', 'FRONTMATTER', 'CONSTAMEND', 'PROCLAMATION', 'REORGPLAN']
        if bill.find("mods:extension/mods:granuleClass", mods_ns).text not in ["PUBLICLAW", "PRIVATELAW", "HCONRES", "SCONRES"]:
            continue

        # Get the title and source URL (used in error messages).
        title_text = bill.find("mods:titleInfo/mods:title", mods_ns).text.replace('""', '"')
        source_url = bill.find("mods:location/mods:url[@displayLabel='Content Detail']", mods_ns).text

        # Bill number
        bill_elements = bill.findall("mods:extension/mods:bill[@priority='primary']", mods_ns)
        if len(bill_elements) == 0:
            logging.error("No bill number identified for '%s' (%s)" % (title_text, source_url))
            continue
        elif len(bill_elements) > 1:
            logging.error("Multiple bill numbers identified for '%s'" % title_text)
            for be in bill_elements:
                logging.error("  -- " + etree.tostring(be).strip())
            logging.error("  @ " + source_url)
            continue
        else:
            bill_congress = bill_elements[0].attrib["congress"]
            bill_type = bill_elements[0].attrib["type"].lower()
            bill_number = bill_elements[0].attrib["number"]
            bill_id = "%s%s-%s" % (bill_type, bill_number, bill_congress)

        # Title
        titles = []
        titles.append({
            "title": title_text,
            "as": "enacted",
            "type": "official",
            "is_for_portion": False,
        })

        # Subject
        descriptor = bill.find("mods:extension/mods:descriptor", mods_ns)
        if descriptor is not None:
            subject = descriptor.text
        else:
            subject = None

        # Committees
        committees = []
        cong_committee = bill.find("mods:extension/mods:congCommittee", mods_ns)
        if cong_committee is not None:
            chambers = {"H": "House", "S": "Senate", "J": "Joint"}
            committee = chambers[cong_committee.attrib["chamber"]] + " " + cong_committee.find("mods:name", mods_ns).text
            committee_info = {
                "committee": committee,
                "activity": [],  # XXX
                "committee_id": utils.committee_names[committee] if committee in utils.committee_names else None,
            }
            committees.append(committee_info)

        # The 'granuleDate' is the enactment date?
        granule_date = bill.find("mods:extension/mods:granuleDate", mods_ns).text

        sources = [{
            "source": "statutes",
            "package_id": package_id,
            "access_id": bill.find("mods:extension/mods:accessId", mods_ns).text,
            "source_url": source_url,
            "volume": bill.find("mods:extension/mods:volume", mods_ns).text,
            "page": bill.find("mods:part[@type='article']/mods:extent[@unit='pages']/mods:start", mods_ns).text,
            "position": bill.find("mods:extension/mods:pagePosition", mods_ns).text,
        }]

        law_elements = bill.findall("mods:extension/mods:law", mods_ns)

        # XXX: If <law> is missing, this assumes it is a concurrent resolution.
        #      This may be a problem if the code is updated to accept joint resolutions for constitutional amendments.
        if (law_elements is None) or (len(law_elements) != 1):
            other_chamber = {"HOUSE": "s", "SENATE": "h"}

            actions = [{
                "type": "vote",
                "vote_type": "vote2",
                "where": other_chamber[bill.find("mods:extension/mods:originChamber", mods_ns).text],
                "result": "pass",  # XXX
                "how": "unknown",  # XXX
                #        "text": "",
                "acted_at": granule_date,  # XXX
                "status": "PASSED:CONCURRENTRES",
                "references": [],  # XXX
            }]
        else:
            law_congress = law_elements[0].attrib["congress"]
            law_number = law_elements[0].attrib["number"]
            law_type = ("private" if (law_elements[0].attrib["isPrivate"] == "true") else "public")

            # Check for typos in the metadata.
            if law_congress != bill_congress:
                logging.error("Congress mismatch for %s%s: %s or %s? (%s)" % (bill_type, bill_number, bill_congress, law_congress, source_url))
                continue

            actions = [{
                "congress": law_congress,
                "number": law_number,
                "type": "enacted",
                "law": law_type,
                "text": "Became %s Law No: %s-%s." % (law_type.capitalize(), law_congress, law_number),
                "acted_at": granule_date,  # XXX
                "status": "ENACTED:SIGNED",  # XXX: Check for overridden vetoes!
                "references": [],  # XXX
            }]

        status, status_date = bill_info.latest_status(actions)

        bill_data = {
            'bill_id': bill_id,
            'bill_type': bill_type,
            'number': bill_number,
            'congress': bill_congress,

            'introduced_at': None,  # XXX
            'sponsor': None,  # XXX
            'cosponsors': [],  # XXX

            'actions': actions,  # XXX
            'history': bill_info.history_from_actions(actions),
            'status': status,
            'status_at': status_date,
            'enacted_as': bill_info.slip_law_from(actions),

            'titles': titles,
            'official_title': bill_info.current_title_for(titles, "official"),
            'short_title': bill_info.current_title_for(titles, "short"),  # XXX
            'popular_title': bill_info.current_title_for(titles, "popular"),  # XXX

            'subjects_top_term': subject,
            'subjects': [],

            'related_bills': [],  # XXX: <associatedBills> usually only lists the current bill.
            'committees': committees,
            'amendments': [],  # XXX

            'sources': sources,
            'updated_at': datetime.datetime.fromtimestamp(time.time()),
        }

        if not options.get('textversions', False):
            bill_info.output_bill(bill_data, options)

        # XXX: Can't use bill_versions.fetch_version() because it depends on fdsys.
        version_code = "enr"
        bill_version_id = "%s%s-%s-%s" % (bill_type, bill_number, bill_congress, version_code)
        bill_version = {
            'bill_version_id': bill_version_id,
            'version_code': version_code,
            'issued_on': status_date,
            'urls': {"pdf": bill.find("mods:location/mods:url[@displayLabel='PDF rendition']", mods_ns).text},
            'sources': sources,
        }
        utils.write(
            json.dumps(bill_version, sort_keys=True, indent=2, default=utils.format_datetime),
            bill_versions.output_for_bill_version(bill_version_id)
        )

        # Process the granule PDF.
        # - Hard-link it into the right place to be seen as bill text.
        # - Run "pdftotext -layout" to convert it to plain text and save it in the bill text location.
        pdf_file = path + "/" + sources[0]["access_id"] + "/document.pdf"
        if os.path.exists(pdf_file):
            dst_path = fdsys.output_for_bill(bill_data["bill_id"], "text-versions/" + version_code, is_data_dot=False)
            if options.get("linkpdf", False):
                os.link(pdf_file, dst_path + "/document.pdf")  # a good idea
            if options.get("extracttext", False):
                logging.error("Running pdftotext on %s..." % pdf_file)
                if subprocess.call(["pdftotext", "-layout", pdf_file, dst_path + "/document.txt"]) != 0:
                    raise Exception("pdftotext failed on %s" % pdf_file)

    return {'ok': True, 'saved': True}



from congress.tasks import utils
import logging
import sys
import os
from datetime import date, datetime
import time
from dateutil.relativedelta import relativedelta
from dateutil.relativedelta import MO
import lxml
import json
import re
import subprocess

from bs4 import BeautifulSoup

from congress.tasks.bills import output_for_bill

# Parsing data from the House' upcoming floor feed, at
# https://docs.house.gov/floor/
#
# This contains data on what bills and draft bills are coming up
# on the floor of the House.
#
# This script will transform the data in the provided XML feed to JSON,
# and download associated documents to disk.
#
# TODO:
#   * Detect and extract any XML files attached to PDFs.
#   * parsing out metadata from any provided XML documents.
#   * handle 'subitems' (e.g. House committee reports)
#
# options:
#   week_of: the date of a Monday of a week to look for. defaults to current week.
#   download: download associated documents and convert PDFs to text


def run(options):
    # accepts yyyymmdd format
    given_week = options.get("week_of", None)
    if given_week is None:
        for_the_weeks = get_mondays_to_scan(options)
    else:
        for_the_weeks = [get_monday_of_week(given_week)]

    # fetch info
    for for_the_week in for_the_weeks:
        run_for_week(for_the_week, options)


def run_for_week(for_the_week, options):
    logging.info(
        "Scraping upcoming bills from docs.house.gov/floor for the week of %s..."
        % for_the_week
    )
    house_floor = fetch_floor_week(for_the_week, options)
    if house_floor is None:
        logging.warn("Nothing posted for the week of %s" % for_the_week)
        return

    output_file = "%s/upcoming_house_floor/%s.json" % (utils.data_dir(), for_the_week)
    output = json.dumps(
        house_floor, sort_keys=True, indent=2, default=utils.format_datetime
    )

    utils.write(output, output_file)
    cb = options.get("callback", None)
    if cb and callable(cb):
        cb(output_file)
    logging.warn(
        "Found %i bills for the week of %s, written to %s"
        % (len(house_floor["upcoming"]), for_the_week, output_file)
    )


# For any week, e.g. https://docs.house.gov/floor/Download.aspx?file=/billsthisweek/20131021/20131021.xml
def fetch_floor_week(for_the_week, options):
    base_url = "https://docs.house.gov/floor/Download.aspx?file=/billsthisweek/"
    week_url = base_url + "%s/%s.xml" % (for_the_week, for_the_week)

    # Turn on 'force' to re-download the schedules, by default, since the content
    # changes frequently and we're scanning weeks that might have 404'd previously
    # when we looked ahead. We leave 'force' off for downloading the file attachments.
    options2 = dict(options)
    if "force" not in options2:
        options2["force"] = True

    body = utils.download(
        week_url, "upcoming_house_floor/%s.xml" % for_the_week, options2
    )
    if "was not found" in body:
        return None
    dom = lxml.etree.fromstring(body)

    # can download the actual attached files to disk, if asked
    download = options.get("download", False)

    # always present at the feed level
    congress = int(dom.xpath("//floorschedule")[0].get("congress-num"))

    # week of this day, e.g. '2013-01-21'
    legislative_day = (
        for_the_week[0:4] + "-" + for_the_week[4:6] + "-" + for_the_week[6:]
    )

    upcoming = []

    for node in dom.xpath("//floorschedule/category/floor-items/floor-item"):
        description = node.xpath("string(floor-text)")

        bill_number = node.xpath("string(legis-num)")

        # TODO: fetch non-bills too
        if not bill_number:
            logging.warn("Skipping item, not a bill: %s" % description)
            continue

        # how is this bill being considered?
        category = next(node.iterancestors("category")).get("type")
        if "suspension" in category:
            consideration = "suspension"
        elif "pursuant" in category:
            consideration = "rule"
        else:
            consideration = "unknown"

        logging.warn("[%s]" % bill_number)

        # todo: establish most recent date from a combo of added, published, updates
        date = date_for(node.get("publish-date"))

        # all items will have this
        bill = {
            "description": description,
            "floor_item_id": node.get("id"),
            "consideration": consideration,
            "published_at": date_for(node.get("publish-date")),
            "added_at": date_for(node.get("add-date")),
        }

        # treat drafts and numbered bills a bit differently
        if "_" in bill_number:
            draft_bill_id = draft_bill_id_for(bill_number, date, congress)
            bill["item_type"] = "draft_bill"
            bill["draft_bill_id"] = draft_bill_id
        else:
            m = re.match(
                "(Concur(ring)? in )?(?P<type>((the )?(Senate|House) Amendments? (with an amendment )?to )+)(?P<bill>.*)",
                bill_number,
                re.I,
            )
            if m:
                amendment_type = m.group("type").split("to")[0]
                if "Senate" in amendment_type and "House" not in amendment_type:
                    bill["item_type"] = "senate_amendment"
                elif "House" in amendment_type and "Senate" not in amendment_type:
                    bill["item_type"] = "house_amendment"
                else:
                    raise ValueError(bill_number)
                bill_number = m.group("bill")

            elif re.match("Conference report to accompany ", bill_number, re.I):
                bill["item_type"] = "conference_report"
                bill_number = bill_number.lower().replace(
                    "conference report to accompany ", ""
                )
            else:
                bill["item_type"] = "bill"

            # In one case we got "H. Res. 497 (H. Rept. 116-125)".
            # Stop at parens.
            bill_number = re.sub(r"\(.*", "", bill_number)

            try:
                bill["bill_id"] = bill_id_for(bill_number.strip(), congress)
            except ValueError:
                logging.error("Could not parse bill from: %s" % bill_number)
                continue

        bill["files"] = []
        for file in node.xpath("files/file"):
            file_url = file.get("doc-url")
            filename = file_url.split("/")[-1]
            file_format = file.get("doc-type").lower()

            logging.warn(
                "\t%s file for %s: %s" % (file_format.upper(), bill_number, filename)
            )

            file_field = {
                "url": file_url,
                "format": file_format,
                "added_at": date_for(file.get("add-date")),
                "published_at": date_for(file.get("publish-date")),
            }

            bill["files"].append(file_field)

            # now try downloading the file to disk and linking it to the data
            if not download:
                continue
            try:
                file_path = "upcoming_house_floor/%s/%s" % (for_the_week, filename)
                try:
                    os.makedirs(
                        os.path.join(utils.data_dir(), os.path.dirname(file_path))
                    )
                except OSError:
                    pass  # directory exists
                options3 = dict(options)
                options3["to_cache"] = False  # put in the actual specified directory
                options3["binary"] = True  # force binary mode, no file escaping
                utils.download(
                    file_url, os.path.join(utils.data_dir(), file_path), options3
                )
                file_field["path"] = file_path
            except IOError:
                logging.error(
                    "Omitting 'path', couldn't download file %s from House floor for the week of %s"
                    % (file_field["url"], for_the_week)
                )
                continue

            # if it's a PDF, convert to text and extract XML
            if file_format == "pdf" and file_path.endswith(".pdf"):
                # extract text
                text_path = file_path.replace(".pdf", ".txt")
                if (
                    subprocess.call(
                        [
                            "pdftotext",
                            "-layout",
                            os.path.join(utils.data_dir(), file_path),
                            os.path.join(utils.data_dir(), text_path),
                        ],
                        universal_newlines=True,
                    )
                    != 0
                ):
                    raise Exception("pdftotext failed on %s" % file_path)
                file_field["text_path"] = text_path

                # extract embedded XML
                for line in subprocess.check_output(
                    ["pdfdetach", "-list", os.path.join(utils.data_dir(), file_path)],
                    universal_newlines=True,
                ).split("\n"):
                    m = re.match(r"(\d+):\s*(.*)", line)
                    if m:
                        attachment_n, attachment_fn = m.groups()
                        if attachment_fn.endswith(".xml"):
                            text_path = file_path.replace(".pdf", ".xml")
                            subprocess.check_call(
                                [
                                    "pdfdetach",
                                    os.path.join(utils.data_dir(), file_path),
                                    "-save",
                                    attachment_n,
                                    "-o",
                                    os.path.join(utils.data_dir(), text_path),
                                ],
                                universal_newlines=True,
                            )
                            file_field["xml_path"] = text_path

        upcoming.append(bill)

        if "bill_id" in bill:
            # Save this bill data to the bill's bill text directory.
            text_data_path = output_for_bill(
                bill["bill_id"],
                os.path.join("text-versions", "dhg-" + bill["floor_item_id"] + ".json"),
                is_data_dot=False,
            )
            try:
                os.makedirs(
                    os.path.join(utils.data_dir(), os.path.dirname(text_data_path))
                )
            except OSError:
                pass  # directory exists
            utils.write(
                json.dumps(
                    bill, sort_keys=True, indent=2, default=utils.format_datetime
                ),
                text_data_path,
            )

    # Create and return the house floor file data.
    house_floor = {
        "congress": congress,
        "week_of": legislative_day,
        "upcoming": upcoming,
    }

    return house_floor


def get_monday_of_week(day_to_get_bills):
    formatted_day = datetime.strptime(day_to_get_bills, "%Y%m%d").date()
    return (formatted_day + relativedelta(weekday=MO(-1))).strftime("%Y%m%d")


# actually go fetch docs.house.gov/floor/ and scrape the download link out of it


def get_latest_monday(options):
    # docs.house.gov always links to the most recent week that isn't in the future.
    url = "https://docs.house.gov/floor/"
    html = utils.download(url, None, options)
    doc = BeautifulSoup(html, features="lxml")

    links = doc.select("a.downloadXML")
    if len(links) != 1:
        utils.admin(
            "There is no docs.house.gov download link --- maybe there are no upcoming bills."
        )
        return None

    link = links[0]
    week = os.path.split(link["href"])[-1].split(".")[0]
    week = datetime.strptime(week, "%Y%m%d").date()

    return week


def get_mondays_to_scan(options):
    # Get the week currently linked on docs.house.gov. If there isn't any (e.g. we are between
    # sessions), just return an empty list of weeks to scan.
    most_recent = get_latest_monday(options)
    if most_recent is None:
        return []

    # Look two weeks into the future too, since when we get to the end of the week the next
    # week's list is sometimes available, and sometimes a week beyond that.
    return [
        (most_recent + relativedelta(days=7 * i)).strftime("%Y%m%d") for i in [0, 1, 2]
    ]


def bill_id_for(bill_number, congress):
    number = bill_number.replace(".", "").replace(" ", "").lower()
    if not re.match(r"^(hr|s|hres|sres|hjres|sjres|hconres|sconres)\d{1,4}$", number):
        raise ValueError(number)
    return "%s-%i" % (number, congress)


def draft_bill_id_for(bill_number, published_at, congress):
    number = bill_number.replace(".", "").replace(" ", "").replace("_", "").lower()
    epoch = time.mktime(published_at.timetuple())
    return "%s%i-%i" % (number, epoch, congress)


def date_for(timestamp):
    if "." not in timestamp:
        strptime_config = "%Y-%m-%dT%H:%M:%S"
    else:
        strptime_config = "%Y-%m-%dT%H:%M:%S.%f"
    return datetime.strptime(timestamp, strptime_config)



import committee_meetings
import lxml.etree
import os
import unittest
import utils

# Parsing the House hearing info


class HearingInfo(unittest.TestCase):

    def test_hearing(self):
        committees = {}
        for c in utils.yaml_load(
                "test/fixtures/committees-current.yaml"):
            committees[c["thomas_id"]] = c
            if "house_committee_id" in c:
                committees[c["house_committee_id"] + "00"] = c
            c["subcommittees"] = dict((s["thomas_id"], s)
                                      for s in c.get("subcommittees", []))

        hearing_xml = "test/fixtures/hearings/sample_hearing.xml"
        file_xml = open(hearing_xml, "r")
        dom = lxml.etree.parse(file_xml)
        test_output = committee_meetings.parse_house_committee_meeting(
            '102252', dom, [], committees, {"debug": False}, None, ["BILLS-113hr4435ih.pdf", "BILLS-113hr4435ih.xml"])
        
#          event_id, dom, existing_meetings, committees, options, witnesses, uploaded_documents
        self.assertEqual(test_output['bill_ids'], ['hr4435-113'])
        self.assertEqual(test_output['chamber'], 'house')
        self.assertEqual(test_output['committee'], 'HSRU')
        self.assertEqual(test_output['congress'], 113)
        self.assertEqual(test_output['house_meeting_type'], 'HMTG')
        self.assertEqual(test_output['meeting_documents'][0][
                         'description'], 'H.R. 4435 (as introduced)')
        self.assertEqual(
            test_output['meeting_documents'][0]['bill_id'],
            'hr4435-113')
        self.assertEqual(
            test_output['meeting_documents'][0]['version_code'], 'ih')
        self.assertEqual(test_output['meeting_documents'][0]['type'], 'BR')
        self.assertEqual(
            test_output['meeting_documents'][0]['urls'], [
                        {'url': 'http://beta.congress.gov/113/bills/hr4435/BILLS-113hr4435ih.pdf', 'file_found': True}, 
                        {'url': 'http://beta.congress.gov/113/bills/hr4435/BILLS-113hr4435ih.xml', 'file_found': True},])
        self.assertEqual(test_output['occurs_at'], '2014-05-19T17:00:00')
        self.assertEqual(test_output['room'], 'CAPITOL H-313')
        self.assertEqual(test_output['subcommittee'], None)
        self.assertEqual(test_output[
                         'topic'], 'H.R. 4435\u2014National Defense Authorization Act for Fiscal Year 2015 [General Debate]; H.R. 4660\u2014Commerce, Justice, Science, and Related Agencies Appropriations Act, 2015')
        self.assertEqual(test_output[
                         'url'], 'http://docs.house.gov/Committee/Calendar/ByEvent.aspx?EventID=102252')

    def test_witnesses(self):
        witness_xml = "test/fixtures/hearings/sample_witness.xml"
        file_xml = open(witness_xml, "r")
        witness_tree = lxml.etree.parse(file_xml)
        uploaded_documents = ["HHRG-113-GO25-Bio-CochraneJ-20140522.pdf"]
        event_id = "102266"
        test_output = committee_meetings.parse_witness_list(witness_tree, uploaded_documents, event_id)["hearing_witness_info"]

        self.assertEqual(test_output[0]['documents'][0]['type'], 'WB')
        self.assertEqual(
            test_output[0]['documents'][0]['description'], 'Cochrane Bio')
        self.assertEqual(test_output[0]['documents'][0]['urls'], [
                         {'url': 'http://docs.house.gov/meetings/GO/GO25/20140522/102266/HHRG-113-GO25-Bio-CochraneJ-20140522.pdf', 'file_found': True}])
        self.assertEqual(test_output[0]['house_event_id'], '102266')
        self.assertEqual(test_output[0]['first_name'], 'James')
        self.assertEqual(test_output[0]['last_name'], 'Cochrane')
        self.assertEqual(
            test_output[0]['position'],
            'Chief Information Officer and Executive Vice President')
        self.assertEqual(test_output[0]['organization'], 'U.S. Postal Service')



import unittest
import bill_info

# parsing various kinds of action text to extract metadata and establish state


def parse_bill_action(line, state, bill_id, title):
    return bill_info.parse_bill_action({"text": line}, state, bill_id, title)


class BillActions(unittest.TestCase):

    def test_veto(self):
        bill_id = "hjres64-111"
        title = "Making further continuing appropriations for fiscal year 2010, and for other purposes."
        state = "PASSED:BILL"
        line = "Vetoed by President."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vetoed")
        self.assertEqual(new_state, "PROV_KILL:VETO")

    def test_pocket_veto(self):
        bill_id = "hr2415-106"
        title = "United Nations Reform Act of 1999"
        state = "PASSED:BILL"
        line = "Pocket Vetoed by President."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vetoed")
        self.assertEqual(new_action['pocket'], "1")
        self.assertEqual(new_state, "VETOED:POCKET")

    def test_reported_from_committee(self):
        bill_id = "s968-112"
        title = "A bill to prevent online threats to economic creativity and theft of intellectual property, and for other purposes."
        state = "REFERRED"
        line = "Committee on the Judiciary. Ordered to be reported with an amendment in the nature of a substitute favorably."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], 'calendar')
        # self.assertEqual(new_action['committee'], "Committee on the Judiciary")
        self.assertEqual(new_state, "REPORTED")

    def test_added_to_calendar(self):
        bill_id = "s968-112"
        title = "A bill to prevent online threats to economic creativity and theft of intellectual property, and for other purposes."
        state = "REPORTED"
        line = "Placed on Senate Legislative Calendar under General Orders. Calendar No. 70."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], 'calendar')
        self.assertEqual(new_action['calendar'], "Senate Legislative")
        self.assertEqual(new_action['under'], "General Orders")
        self.assertEqual(new_action['number'], "70")
        self.assertEqual(new_state, None)

    def test_enacted_as_public_law(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "ENACTED:SIGNED"
        line = "Became Public Law No: 111-148."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "enacted")
        self.assertEqual(new_action['congress'], "111")
        self.assertEqual(new_action['number'], "148")
        self.assertEqual(new_action['law'], "public")

    def test_cleared_for_whitehouse(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASSED:BILL"
        line = "Cleared for White House."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        # should not be marked as presented to president, since it hasn't been yet
        # self.assertEqual(new_action['type'], 'action')

    def test_presented_to_president(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASSED:BILL"
        line = "Presented to President."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], 'topresident')

    def test_signed_by_president(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASSED:BILL"
        line = "Signed by President."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], 'signed')

    # voting tests

    def test_vote_normal_roll(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "INTRODUCED"
        line = "On motion to suspend the rules and pass the bill Agreed to by the Yeas and Nays: (2/3 required): 416 - 0 (Roll no. 768)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")
        self.assertEqual(new_action['roll'], "768")

        self.assertEqual(new_state, "PASS_OVER:HOUSE")

    def test_vote_normal_roll_second(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASS_OVER:HOUSE"
        line = "Passed Senate with an amendment and an amendment to the Title by Yea-Nay Vote. 60 - 39. Record Vote Number: 396."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote2")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")
        self.assertEqual(new_action['roll'], "396")

        self.assertEqual(new_state, "PASS_BACK:SENATE")

    def test_cloture_vote_verbose(self):
        bill_id = "s1982-113"
        title = "Comprehensive Veterans Health and Benefits and Military Retirement Pay Restoration Act of 2014"
        line = "Cloture motion on the motion to proceed to the measure invoked in Senate by Yea-Nay Vote. 99 - 0. Record Vote Number: 44."
        state = "REPORTED"

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "vote-aux")
        self.assertEqual(new_action['vote_type'], "cloture")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")
        self.assertEqual(new_action['roll'], "44")

        self.assertEqual(new_state, None)

    def test_vote_roll_pingpong(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASS_BACK:SENATE"
        line = "On motion that the House agree to the Senate amendments Agreed to by recorded vote: 219 - 212 (Roll no. 165)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['roll'], "165")
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "pingpong")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")

    def test_vote_cloture(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASS_OVER:HOUSE"  # should not change
        line = "Cloture on the motion to proceed to the bill invoked in Senate by Yea-Nay Vote. 60 - 39. Record Vote Number: 353."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['roll'], "353")
        self.assertEqual(new_action['type'], "vote-aux")
        self.assertEqual(new_action['vote_type'], "cloture")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")

        self.assertEqual(new_state, None)  # unchanged

    def test_vote_cloture_2(self):
        bill_id = "hr3590-111"
        title = "An act entitled The Patient Protection and Affordable Care Act."
        state = "PASS_OVER:HOUSE"  # should not change
        line = "Cloture invoked in Senate by Yea-Nay Vote. 60 - 39. Record Vote Number: 395."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['roll'], "395")
        self.assertEqual(new_action['type'], "vote-aux")
        self.assertEqual(new_action['vote_type'], "cloture")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action['how'], "roll")
        self.assertEqual(new_action['result'], "pass")

        self.assertEqual(new_state, None)  # unchanged

    # not sure whether to include votes that are on process, not passage or cloture

    # def test_vote_process_voice_senate(self):
    #   bill_id = "hr3590-111"
    #   title = "An act entitled The Patient Protection and Affordable Care Act."
    # state = "PASS_OVER:HOUSE" # should not change
    #   line = "Motion to proceed to consideration of measure agreed to in Senate by Unanimous Consent."

    #   new_action, new_state = parse_bill_action(line, state, bill_id, title)

    #   self.assertEqual(new_action['type'], 'vote')
    #   self.assertEqual(new_action['vote_type'], 'other')
    #   self.assertEqual(new_action['how'], 'Unanimous Consent')
    #   self.assertEqual(new_action['where'], 's')
    #   self.assertEqual(new_action['result'], 'pass')
    #   self.assertEqual(new_state, None)

    # def test_vote_commit_roll_failure(self):
    #   bill_id = "hr3590-111"
    #   title = "An act entitled The Patient Protection and Affordable Care Act."
    # state = "PASS_OVER:HOUSE" # should not change
    #   line = "Motion by Senator McCain to commit to Senate Committee on Finance under the order of 12/2/2009, not having achieved 60 votes in the affirmative, the motion was rejected in Senate by Yea-Nay Vote. 42 - 58. Record Vote Number: 358."

    #   new_action, new_state = parse_bill_action(line, state, bill_id, title)

    #   self.assertEqual(new_action['type'], 'vote')
    #   self.assertEqual(new_action['vote_type'], 'other')
    #   self.assertEqual(new_action['how'], 'roll')
    #   self.assertEqual(new_action['where'], 's')
    #   self.assertEqual(new_action['result'], 'fail')
    #   self.assertEqual(new_action['roll'], "358")
    #   self.assertEqual(new_state, None)

    # def test_vote_motion_conference(self):
    #   bill_id = "hr3630-112"
    #   title = "A bill to extend the payroll tax holiday, unemployment compensation, Medicare physician payment, provide for the consideration of the Keystone XL pipeline, and for other purposes."
    #   state = "PASS_BACK:SENATE"
    #   line = "On motion that the House disagree to the Senate amendments, and request a conference Agreed to by the Yeas and Nays: 229 - 193 (Roll no. 946)."

    #   new_action, new_state = parse_bill_action(line, state, bill_id, title)

    # self.assertEqual(new_action['type'], 'vote')
    # self.assertEqual(new_action['vote_type'], 'other')
    # self.assertEqual(new_action['how'], 'roll')
    # self.assertEqual(new_action['where'], 'h')
    # self.assertEqual(new_action['result'], 'pass')
    # self.assertEqual(new_action['roll'], "946")
    #   self.assertEqual(new_state, None)

    # def test_vote_motion_instruct_conferees(self):
    #   bill_id = "hr3630-112"
    #   title = "A bill to extend the payroll tax holiday, unemployment compensation, Medicare physician payment, provide for the consideration of the Keystone XL pipeline, and for other purposes."
    #   state = "PASS_BACK:SENATE"
    #   line = "On motion that the House instruct conferees Agreed to by the Yeas and Nays: 397 - 16 (Roll no. 9)."

    #   new_action, new_state = parse_bill_action(line, state, bill_id, title)

    # self.assertEqual(new_action['type'], 'vote')
    # self.assertEqual(new_action['vote_type'], 'other')
    # self.assertEqual(new_action['how'], 'roll')
    # self.assertEqual(new_action['where'], 'h')
    # self.assertEqual(new_action['result'], 'pass')
    # self.assertEqual(new_action['roll'], "9")
    #   self.assertEqual(new_state, None)

    def test_vote_conference_report_house_pass(self):
        bill_id = "hr3630-112"
        title = "A bill to extend the payroll tax holiday, unemployment compensation, Medicare physician payment, provide for the consideration of the Keystone XL pipeline, and for other purposes."
        state = "PASS_BACK:SENATE"
        line = "On agreeing to the conference report Agreed to by the Yeas and Nays: 293 - 132 (Roll no. 72)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], 'vote')
        self.assertEqual(new_action['vote_type'], 'conference')
        self.assertEqual(new_action['how'], 'roll')
        self.assertEqual(new_action['where'], 'h')
        self.assertEqual(new_action['result'], 'pass')
        self.assertEqual(new_action['roll'], "72")
        self.assertEqual(new_state, 'CONFERENCE:PASSED:HOUSE')

    def test_vote_conference_report_senate_pass(self):
        bill_id = "hr3630-112"
        title = "A bill to extend the payroll tax holiday, unemployment compensation, Medicare physician payment, provide for the consideration of the Keystone XL pipeline, and for other purposes."
        state = "CONFERENCE:PASSED:HOUSE"
        line = "Senate agreed to conference report by Yea-Nay Vote. 60 - 36. Record Vote Number: 22."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], 'vote')
        self.assertEqual(new_action['vote_type'], 'conference')
        self.assertEqual(new_action['how'], 'roll')
        self.assertEqual(new_action['where'], 's')
        self.assertEqual(new_action['result'], 'pass')
        self.assertEqual(new_action['roll'], "22")
        self.assertEqual(new_state, 'PASSED:BILL')

    def test_vote_veto_override_fail(self):
        bill_id = "hjres64-111"
        title = "Making further continuing appropriations for fiscal year 2010, and for other purposes."
        state = "PROV_KILL:VETO"
        line = "On passage, the objections of the President to the contrary notwithstanding Failed by the Yeas and Nays: (2/3 required): 143 - 245, 1 Present (Roll no. 2).On passage, the objections of the President to the contrary notwithstanding Failed by the Yeas and Nays: (2/3 required): 143 - 245, 1 Present (Roll no. 2)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "override")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "fail")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action["roll"], "2")
        self.assertEqual(new_state, "VETOED:OVERRIDE_FAIL_ORIGINATING:HOUSE")

    def test_veto_override_success_once(self):
        bill_id = "hr6331-110"
        title = "Medicare Improvements for Patients and Providers Act of 2008"
        state = "PROV_KILL:VETO"
        line = "Two-thirds of the Members present having voted in the affirmative the bill is passed, Passed by the Yeas and Nays: (2/3 required): 383 - 41 (Roll no. 491)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "override")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action["roll"], "491")
        self.assertEqual(new_state, "VETOED:OVERRIDE_PASS_OVER:HOUSE")

    def test_veto_override_success_twice(self):
        bill_id = "hr6331-110"
        title = "Medicare Improvements for Patients and Providers Act of 2008"
        state = "VETOED:OVERRIDE_PASS_OVER:HOUSE"
        line = "Passed Senate over veto by Yea-Nay Vote. 70 - 26. Record Vote Number: 177."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "override")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action["roll"], "177")
        # self.assertEqual(new_state, "VETOED:OVERRIDE_COMPLETE:SENATE")

    # Fictional bill, no constitutional amendment passed by both Houses
    # in the THOMAS era (1973-present).
    # The 26th was passed by Congress in 1971, 27th passed by Congress in 1789.
    # The line here is taken from hjres10-109, when the House passed a
    # flag burning amendment. (A separate version later failed the Senate by one vote.)
    def test_passed_constitutional_amendment(self):
        bill_id = "sjres64-1000"
        title = "Proposing an amendment to the Constitution of the United States authorizing the Congress to prohibit the physical desecration of the flag of the United States."
        state = "PASS_OVER:SENATE"
        line = "On passage Passed by the Yeas and Nays: (2/3 required): 286 - 130 (Roll no. 296)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote2")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action["roll"], "296")
        self.assertEqual(new_state, "PASSED:CONSTAMEND")

    def test_passed_concurrent_resolution(self):
        bill_id = "hconres74-112"
        title = "Providing for a joint session of Congress to receive a message from the President."
        state = "PASS_OVER:HOUSE"
        line = "Received in the Senate, considered, and agreed to without amendment by Unanimous Consent."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote2")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "by Unanimous Consent")
        self.assertEqual(new_state, "PASSED:CONCURRENTRES")

    def test_passed_simple_resolution_house(self):
        bill_id = "hres9-112"
        title = "Instructing certain committees to report legislation replacing the job-killing health care law."
        state = "REPORTED"
        line = "On agreeing to the resolution, as amended Agreed to by the Yeas and Nays: 253 - 175 (Roll no. 16)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action['roll'], "16")
        self.assertEqual(new_state, "PASSED:SIMPLERES")

    def test_passed_simple_resolution_senate(self):
        bill_id = "sres484-112"
        title = "A resolution designating June 7, 2012, as \"National Hunger Awareness Day\"."
        state = "REPORTED"
        line = "Submitted in the Senate, considered, and agreed to without amendment and with a preamble by Unanimous Consent."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "by Unanimous Consent")
        self.assertEqual(new_state, "PASSED:SIMPLERES")

    def test_failed_simple_resolution_senate(self):
        bill_id = "sres5-113"
        title = "A resolution amending the Standing Rules of the Senate to provide for cloture to be invoked with less than a three-fifths majority after additional debate."
        state = "INTRODUCED"
        line = "Disagreed to in Senate by Voice Vote."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "s")
        self.assertEqual(new_action["result"], "fail")
        self.assertEqual(new_action["how"], "by Voice Vote")
        self.assertEqual(new_state, "FAIL:ORIGINATING:SENATE")

    def test_failed_suspension_vote(self):
        bill_id = "hr1954-112"
        title = "To implement the President's request to increase the statutory limit on the public debt."
        state = "REFERRED"
        line = "On motion to suspend the rules and pass the bill Failed by the Yeas and Nays: (2/3 required): 97 - 318, 7 Present (Roll no. 379)."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "fail")
        self.assertEqual(new_action["how"], "roll")
        self.assertEqual(new_action['roll'], "379")
        self.assertEqual(new_state, "PROV_KILL:SUSPENSIONFAILED")

    def test_passed_by_special_rule(self):
        bill_id = "hres240-109"
        title = "Amending the Rules of the House of Representatives to reinstate certain provisions of the rules relating to procedures of the Committee on Standards of Official Conduct to the form in which those provisions existed at the close of the 108th Congress."
        state = "INTRODUCED"
        line = "Passed House pursuant to H. Res. 241."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)
        self.assertEqual(new_action['type'], "vote")
        self.assertEqual(new_action['vote_type'], "vote")
        self.assertEqual(new_action['where'], "h")
        self.assertEqual(new_action["result"], "pass")
        self.assertEqual(new_action["how"], "by special rule")
        self.assertEqual(new_state, "PASSED:SIMPLERES")

        self.assertEqual(new_action['bill_ids'], ["hres241-109"])

    def test_referral_committee(self):
        bill_id = "hr547-113"
        title = "To provide for the establishment of a border protection strategy for the international land borders of the United States, to address the ecological and environmental impacts of border security infrastructure, measures, and activities along the international land borders of the United States, and for other purposes."
        state = "INTRODUCED"
        line = "Referred to the Committee on Homeland Security, and in addition to the Committees on Armed Services, Agriculture, and Natural Resources, for a period to be subsequently determined by the Speaker, in each case for consideration of such provisions as fall within the jurisdiction of the committee concerned."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "referral")
        self.assertEqual(new_state, "REFERRED")

    def test_referral_subcommittee(self):
        bill_id = "hr547-113"
        title = "To provide for the establishment of a border protection strategy for the international land borders of the United States, to address the ecological and environmental impacts of border security infrastructure, measures, and activities along the international land borders of the United States, and for other purposes."
        state = "INTRODUCED"
        line = "Referred to the Subcommittee Indian and Alaska Native Affairs."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "referral")
        self.assertEqual(new_state, "REFERRED")

    def test_hearings_held(self):
        bill_id = "s54-113"
        title = "A bill to increase public safety by punishing and deterring firearms trafficking."
        state = "REFERRED"
        line = "Committee on the Judiciary Subcommittee on the Constitution, Civil Rights and Human Rights. Hearings held."

        new_action, new_state = parse_bill_action(line, state, bill_id, title)

        self.assertEqual(new_action['type'], "hearings")
        # self.assertEqual(new_action['committees'], "Committee on the Judiciary Subcommittee on the Constitution, Civil Rights and Human Rights")
        self.assertEqual(new_state, None)  # did not change state



from congress.tasks import utils
import logging
import re
import json
from datetime import datetime
from lxml import etree
import time
from lxml.html import fromstring

# can be run on its own, just require a nomination_id (e.g. PN2094-112)


def run(options):
	nomination_id = options.get('nomination_id', None)

	if nomination_id:
		result = fetch_nomination(nomination_id, options)
		logging.warn("\n%s" % result)
	else:
		logging.error("To run this task directly, supply a bill_id.")

# download and cache page for nomination


def fetch_nomination(nomination_id, options={}):
	logging.info("\n[%s] Fetching..." % nomination_id)

	# fetch committee name map, if it doesn't already exist
	nomination_type, number, congress = utils.split_nomination_id(nomination_id)
	if not number:
		return {'saved': False, 'ok': False, 'reason': "Couldn't parse %s" % nomination_id}

	if not utils.committee_names:
		utils.fetch_committee_names(congress, options)

	# fetch bill details body
	body = utils.download(
		nomination_url_for(nomination_id),
		nomination_cache_for(nomination_id, "information.html"), options)

	if not body:
		return {'saved': False, 'ok': False, 'reason': "failed to download"}

	if options.get("download_only", False):
		return {'saved': False, 'ok': True, 'reason': "requested download only"}

	# TODO:
	#   detect group nominations, particularly for military promotions
	#   detect when a group nomination is split into subnominations
	#
	# Also, the splitting process is nonsense:
	# http://thomas.loc.gov/home/PN/split.htm

	if "split into two or more parts" in body:
		return {'saved': False, 'ok': True, 'reason': 'was split'}

	nomination = parse_nomination(nomination_id, body, options)
	output_nomination(nomination, options)
	return {'ok': True, 'saved': True}


def parse_nomination(nomination_id, body, options):
	nomination_type, number, congress = utils.split_nomination_id(nomination_id)

	# remove (and store) comments, which contain some info for the nomination
	# but also mess up the parser
	facts = re.findall("<!--(.+?)-->", body)
	body = re.sub("<!--.+?-->", "", body)

	doc = fromstring(body)

	# get rid of centered bold labels, they screw stuff up,
	# e.g. agency names on PN1375-113
	body = re.sub(re.compile("<div align=\"center\">.+?</div>", re.M), "", body)
	for elem in doc.xpath('//div[@align="center"]'):
		elem.getparent().remove(elem)

	committee_names = []
	committees = []

	info = {
		'nomination_id': nomination_id, 'actions': []
	}

	# the markup on these pages is a disaster, so we're going to use a heuristic based on boldface, inline tags followed by text
	for pair in doc.xpath('//span[@class="elabel"]|//strong'):
		if pair.tail:
			text = pair.text or pair.text_content()
			label, data = text.replace(':', '').strip(), pair.tail.strip()

			# handle actions separately
			if label.split(" ")[-1] == "Action":
				pieces = re.split("\s+\-\s+", data)

				location = label.split(" ")[0].lower()

				# use 'acted_at', even though it's always a date, to be consistent
				# with acted_at field on bills and amendments
				acted_at = datetime.strptime(pieces[0], "%B %d, %Y").strftime("%Y-%m-%d")

				# join rest back together (in case action itself has a hyphen)
				text = str.join(" - ", pieces[1:len(pieces)])

				info['actions'].append({
					"type": "action",
					"location": location,
					"acted_at": acted_at,
					"text": text
				})

			else:
				# let's handle these cases one by one
				if label == "Organization":
					info["organization"] = data

				elif label == "Control Number":
					# this doesn't seem useful
					pass

				elif label.lower() == "referred to":
					committee_names.append(data)

				elif label == "Reported by":
					info["reported_by"] = data

				elif label == "Nomination":
					# sanity check - verify nomination_id matches
					if nomination_id != data:
						raise Exception("Whoa! Mismatched nomination ID.")

				elif label == "Date Received":
					# Note: Will break with the 1000th congress in year 3789
					match = re.search("(\d{2,3})[stndhr]{2}", data)
					if match:
						info["congress"] = int(match.group(1))
					else:
						raise Exception("Choked, couldn't find Congress in \"%s\"" % data)

					# Doc format is: "January 04, 1995 (104th Congress)"
					info["received_on"] = datetime.strptime(data.split(" (")[0], "%B %d, %Y").strftime("%Y-%m-%d")

				elif label == "Nominee":

					# ignore any vice suffix
					name = data.split(", vice")[0]

					try:
						name = re.search("(.+?),", name).groups()[0]
					except Exception as e:
						raise Exception("Couldn't parse nominee entry: %s" % name)

					# Some begin "One nomination,...", so 'List of Nominees' will get it
					if "nomination" in name:
						pass

					# and grab the state and position out of the comment facts
					if facts[-5]:
						position = facts[-5]
					else:
						raise Exception("Couldn't find the position in the comments.")

					info["nominees"] = [{
						"name": name,
						"position": position,
						"state": facts[-6][2:]
					}]

				elif label.lower() == "nominees":
					pass

				elif label.lower() == "authority date":
					pass

				elif label.lower() == "list of nominees":
					# step through each sibling, collecting each br's stripped tail for names as we go
					# stop when we get to a strong or span (next label)
					nominees = []

					current_position = None
					for sibling in pair.itersiblings():
						if sibling.tag == "br":
							if sibling.tail:
								name = sibling.tail.strip()
								if (name[0:5].lower() == "to be"):
									current_position = name[6:].strip()
								elif name:
									nominees.append({
										"name": sibling.tail.strip(),
										"position": current_position
									})
						elif (sibling.tag == "strong") or (sibling.tag == "span"):
							break

					info["nominees"] = nominees

				else:
					# choke, I think we handle all of them now
					raise Exception("Unrecognized label: %s" % label)

	if not info.get("received_on", None):
		raise Exception("Choked, couldn't find received date.")

	if not info.get("nominees", None):
		raise Exception("Choked, couldn't find nominee info.")

	# try to normalize committee name to an ID
	# choke if it doesn't work - the names should match up.
	for committee_name in committee_names:
		committee_id = utils.committee_names[committee_name]
		committees.append(committee_id)
	info["referred_to"] = committees
	info["referred_to_names"] = committee_names

	return info

# directory helpers


def output_for_nomination(nomination_id, format):
	nomination_type, number, congress = utils.split_nomination_id(nomination_id)
	return "%s/%s/nominations/%s/%s" % (utils.data_dir(), congress, number, "data.%s" % format)


def nomination_url_for(nomination_id):
	nomination_type, number, congress = utils.split_nomination_id(nomination_id)

	# numbers can be either of the form "63" or "64-01"
	number_pieces = number.split("-")
	if len(number_pieces) == 1:
		number_pieces.append("00")
	url_number = "%05d%s" % (int(number_pieces[0]), number_pieces[1])

	return "http://thomas.loc.gov/cgi-bin/ntquery/z?nomis:%03d%s%s:/" % (int(congress), nomination_type.upper(), url_number)


def nomination_cache_for(nomination_id, file):
	nomination_type, number, congress = utils.split_nomination_id(nomination_id)
	return "%s/nominations/%s/%s" % (congress, number, file)


def output_nomination(nomination, options):
	logging.info("[%s] Writing to disk..." % nomination['nomination_id'])

	# output JSON - so easy!
	utils.write(
		json.dumps(nomination, sort_keys=True, indent=2, default=utils.format_datetime),
		output_for_nomination(nomination['nomination_id'], "json")
	)



from congress.tasks import utils
import logging
import re
import json
from lxml import etree
import copy
import datetime


def create_govtrack_xml(bill, options):
    govtrack_type_codes = {
        "hr": "h",
        "s": "s",
        "hres": "hr",
        "sres": "sr",
        "hjres": "hj",
        "sjres": "sj",
        "hconres": "hc",
        "sconres": "sc",
    }
    root = etree.Element("bill")
    root.set("session", bill["congress"])
    root.set("type", govtrack_type_codes[bill["bill_type"]])
    root.set("number", bill["number"])
    root.set("updated", utils.format_datetime(bill["updated_at"]))

    def make_node(parent, tag, text, **attrs):
        if options.get("govtrack", False):
            # Rewrite bioguide_id attributes as just id with GovTrack person IDs.
            attrs2 = {}
            for k, v in attrs.items():
                if v:
                    if k == "bioguide_id":
                        # remap "bioguide_id" attributes to govtrack "id"
                        k = "id"
                        v = str(
                            utils.translate_legislator_id("bioguide", v, "govtrack")
                        )
                    if k == "thomas_id":
                        # remap "thomas_id" attributes to govtrack "id"
                        k = "id"
                        v = str(utils.translate_legislator_id("thomas", v, "govtrack"))
                    attrs2[k] = v
            attrs = attrs2

        return utils.make_node(parent, tag, text, **attrs)

    # for American Memory Century of Lawmaking bills...
    for source in bill.get("sources", []):
        n = make_node(root, "source", "")
        for k, v in sorted(source.items()):
            if k == "source":
                n.text = v
            elif k == "source_url":
                n.set("url", v)
            else:
                n.set(k, str(v))
    if "original_bill_number" in bill:
        make_node(root, "bill-number", bill["original_bill_number"])

    make_node(root, "state", bill["status"], datetime=bill["status_at"])
    old_status = make_node(root, "status", None)
    make_node(
        old_status,
        "introduced" if bill["status"] in ("INTRODUCED", "REFERRED") else "unknown",
        None,
        datetime=bill["status_at"],
    )  # dummy for the sake of comparison

    make_node(root, "introduced", None, datetime=bill["introduced_at"])
    titles = make_node(root, "titles", None)
    for title in bill["titles"]:
        n = make_node(titles, "title", title["title"])
        n.set("type", title["type"])
        if title["as"]:
            n.set("as", title["as"])
        if title["textVersionCode"]:
            n.set("textVersionCode", title["textVersionCode"])
        if title["is_for_portion"]:
            n.set("partial", "1")

    def get_legislator_id_attr(p):
        if "bioguide_id" in p:
            return {"bioguide_id": p["bioguide_id"]}
        if "thomas_id" in p:
            return {"thomas_id": p["thomas_id"]}
        return {}

    if bill["sponsor"]:
        # TODO: Sponsored by committee?
        make_node(root, "sponsor", None, **get_legislator_id_attr(bill["sponsor"]))
    else:
        make_node(root, "sponsor", None)

    cosponsors = make_node(root, "cosponsors", None)
    for cosp in bill["cosponsors"]:
        n = make_node(cosponsors, "cosponsor", None, **get_legislator_id_attr(cosp))
        if cosp["sponsored_at"]:
            n.set("joined", cosp["sponsored_at"])
        if cosp.get(
            "withdrawn_at"
        ):  # no longer present in GPO BILLSTATUS XML schema 3.0.0
            n.set("withdrawn", cosp["withdrawn_at"])

    actions = make_node(root, "actions", None)
    for action in bill["actions"]:
        a = make_node(
            actions,
            (
                action["type"]
                if action["type"]
                in (
                    "vote",
                    "vote-aux",
                    "calendar",
                    "topresident",
                    "signed",
                    "enacted",
                    "vetoed",
                )
                else "action"
            ),
            None,
            datetime=action["acted_at"],
        )
        if action.get("status"):
            a.set("state", action["status"])
        if action["type"] in ("vote", "vote-aux"):
            a.clear()  # re-insert date between some of these attributes
            a.set("how", action["how"])
            a.set("type", action["vote_type"])
            if action.get("roll") != None:
                a.set("roll", action["roll"])
            a.set("datetime", utils.format_datetime(action["acted_at"]))
            a.set("where", action["where"])
            a.set("result", action["result"])
            if action.get("suspension"):
                a.set("suspension", "1")
            if action.get("status"):
                a.set("state", action["status"])
        if action["type"] == "calendar" and "calendar" in action:
            a.set("calendar", action["calendar"])
            if action["under"]:
                a.set("under", action["under"])
            if action["number"]:
                a.set("number", action["number"])
        if action["type"] == "enacted":
            a.clear()  # re-insert date between some of these attributes
            a.set("number", "%s-%s" % (bill["congress"], action["number"]))
            a.set("type", action["law"])
            a.set("datetime", utils.format_datetime(action["acted_at"]))
            if action.get("status"):
                a.set("state", action["status"])
        if action["type"] == "vetoed":
            if action.get("pocket"):
                a.set("pocket", "1")
        if action.get("text"):
            make_node(a, "text", action["text"])
        if action.get("in_committee"):
            make_node(a, "committee", None, name=action["in_committee"])
        for cr in action["references"]:
            make_node(a, "reference", None, ref=cr["reference"], label=cr["type"])

    committees = make_node(root, "committees", None)
    for cmt in bill["committees"]:
        make_node(
            committees,
            "committee",
            None,
            code=(
                (cmt["committee_id"] + cmt["subcommittee_id"])
                if cmt.get("subcommittee_id", None)
                else cmt["committee_id"]
            ),
            name=cmt["committee"],
            subcommittee=(
                cmt.get("subcommittee").replace("Subcommittee on ", "")
                if cmt.get("subcommittee")
                else ""
            ),
            activity=", ".join(c.title() for c in cmt["activity"]),
        )

    relatedbills = make_node(root, "relatedbills", None)
    for rb in bill["related_bills"]:
        if rb["type"] == "bill":
            rb_bill_type, rb_number, rb_congress = utils.split_bill_id(rb["bill_id"])
            make_node(
                relatedbills,
                "bill",
                None,
                session=rb_congress,
                type=govtrack_type_codes[rb_bill_type],
                number=rb_number,
                relation="unknown" if rb["reason"] == "related" else rb["reason"],
            )

    subjects = make_node(root, "subjects", None)
    if bill["subjects_top_term"]:
        make_node(subjects, "term", None, name=bill["subjects_top_term"])
    for s in bill["subjects"]:
        if s != bill["subjects_top_term"]:
            make_node(subjects, "term", None, name=s)

    amendments = make_node(root, "amendments", None)
    for amd in bill["amendments"]:
        make_node(
            amendments, "amendment", None, number=amd["chamber"] + str(amd["number"])
        )

    if bill.get("summary"):
        make_node(
            root,
            "summary",
            bill["summary"]["text"],
            date=bill["summary"]["date"],
            status=bill["summary"]["as"],
        )

    if bill.get("committee_reports"):
        committee_reports = make_node(root, "committee-reports", None)
        for report in bill.get("committee_reports", []):
            make_node(committee_reports, "report", report)

    return etree.tostring(root, pretty_print=True)


def sponsor_for(sponsor_dict):
    if sponsor_dict is None:
        # TODO: This can hopefully be removed. In testing s414-113
        # was missing sponsor data. But all bills have a sponsor?
        return None

    # TODO: Don't do regex matching here. Find another way.
    m = re.match(
        r"(?P<title>(Rep\.|Sen\.|Del\.|Resident Commissioner|Rescom\.)) (?P<name>.*?) +\[(?P<party>[DRIL])-(?P<state>[A-Z][A-Z])(-(?P<district>\d{1,2}|At Large|None))?\]$",
        sponsor_dict["fullName"],
    )

    if not m:
        raise ValueError(sponsor_dict)

    return {
        "title": m.group("title"),
        "name": m.group(
            "name"
        ),  # the firstName, middleName, lastName fields have inconsistent capitalization - some are all uppercase
        "state": sponsor_dict["state"],
        "district": sponsor_dict.get("district"),  # missing for senators
        #'party': m.group('party'),
        "bioguide_id": sponsor_dict["bioguideId"],
        "type": "person",
    }


def summary_for(summaries):
    # Some bills are missing the summaries entirely?
    if summaries is None:
        return None

    # Take the most recent summary, by looking at the lexicographically last updateDate.
    summary = sorted(summaries, key=lambda s: s["updateDate"])[-1]

    # Congress 118 and after wrap 'text' inside of a 'cdata' field
    text = summary.get("text", "")
    if not text:
        cdata = summary.get("cdata", "")
        if isinstance(cdata, dict):
            text = cdata.get("text", "")

    # Build dict.
    return {
        "date": summary["updateDate"],
        "as": summary["actionDesc"],
        "asOf": summary["actionDate"],
        "text": strip_tags(text),
    }


def strip_tags(text):
    # Preserve paragraph breaks. Convert closing p tags (and surrounding whitespace) into two newlines. Strip trailing whitespace
    text = re.sub("\s*</\s*p\s*>\s*", "\n\n", text).strip()

    # naive stripping of tags, should work okay in this limited context
    text = re.sub("<[^>]+>", "", text)

    # compress and strip whitespace artifacts, except for the paragraph breaks
    text = re.sub("[ \t\r\f\v]{2,}", " ", text).strip()

    # Replace HTML entities with characters.
    text = utils.unescape(text)

    return text


def committees_for(committee_list):
    if committee_list is None:
        return []

    committee_list = committee_list["item"]

    activity_text_map = {
        "Referred to": ["referral"],
        "Hearings by": ["hearings"],
        "Markup by": ["markup"],
        "Reported by": ["reporting"],
        "Discharged from": ["discharged"],
        "Reported original measure": ["origin", "reporting"],
    }

    def fix_subcommittee_name(name):
        return re.sub(
            "(.*) Subcommittee$", lambda m: "Subcommittee on " + m.group(1), name
        )

    def get_activitiy_list(item):
        if not item.get("activities"):
            return []
        return sum(
            [
                activity_text_map.get(i["name"], [i["name"]])
                for i in item["activities"]["item"]
            ],
            [],
        )

    def fixup_committee_name(name):
        # Preserve backwards compatiblity.
        if name == "House House Administration":
            return "House Administration"
        return name

    def build_dict(item):
        committee_dict = {
            "activity": get_activitiy_list(item),
            "committee": fixup_committee_name(
                item["chamber"] + " " + re.sub(" Committee$", "", item["name"])
            ),
            "committee_id": item["systemCode"][0:-2].upper(),
        }

        subcommittees_list = []
        if "subcommittees" in item and item["subcommittees"] is not None:
            for subcommittee in item["subcommittees"]["item"]:
                subcommittee_dict = copy.deepcopy(committee_dict)
                subcommittee_dict.update(
                    {
                        "subcommittee": fix_subcommittee_name(subcommittee["name"]),
                        "subcommittee_id": subcommittee["systemCode"][-2:],
                        "activity": get_activitiy_list(subcommittee),
                    }
                )
                subcommittees_list.append(subcommittee_dict)

        return [committee_dict] + subcommittees_list

    return sum([build_dict(committee) for committee in committee_list], [])


def titles_for(title_list):
    def build_dict(item):

        full_type = item["titleType"]
        is_for_portion = False

        # "Official Titles as Introduced", "Short Titles on Conference report"
        splits = re.split(" as | on ", full_type, 1)
        if len(splits) == 2:
            title_type, state = splits

            if state.endswith(" for portions of this bill"):
                is_for_portion = True
                state = state.replace(" for portions of this bill", "")

            state = state.replace(":", "").lower()
        else:
            title_type, state = full_type, None

        if "Popular Title" in title_type:
            title_type = "popular"
        elif "Short Title" in title_type:
            title_type = "short"
        elif "Official Title" in title_type:
            title_type = "official"
        elif "Display Title" in title_type:
            title_type = "display"
        elif title_type == "Non-bill-report":
            # TODO: What kind of title is this? Maybe assign
            # a better title_type code once we know.
            title_type = "nonbillreport"
        else:
            raise Exception("Unknown title type: " + title_type)

        return {
            "title": item["title"],
            "is_for_portion": is_for_portion,
            "as": state,
            "textVersionCode": item.get("TextVersionCode"),
            "type": title_type,
        }

    titles = [build_dict(title) for title in title_list]

    # THOMAS used to give us the titles in a particular order:
    #  short as introduced
    #  short as introduced (for portion)
    #  short as some later stage
    #  short as some later stage (for portion)
    #  official as introduced
    #  official as some later stage
    # The "as" stages (introduced, etc.) were in the order in which actions
    # actually occurred. This was handy because to get the current title for
    # a bill, you need to know which action type was most recent. The new
    # order is reverse-chronological, so we have to turn the order around
    # for backwards compatibility. Rather than do a simple .reverse(), I'm
    # adding an explicit sort order here which gets very close to the THOMAS
    # order.
    # Unfortunately this can no longer be relied on because the new bulk
    # data has the "as" stages sometimes in the wrong order: The "reported to
    # senate" status for House bills seems to be consistently out of place.
    titles_copy = list(titles)  # clone before beginning sort

    def first_index_of(**kwargs):
        for i, title in enumerate(titles_copy):
            for k, v in kwargs.items():
                k = k.replace("_", "")
                if title.get(k) != v:
                    break
            else:
                # break not called --- all match
                return i

    titles.sort(
        key=lambda title: (
            # keep the same 'short', 'official', 'display' order intact
            first_index_of(type=title["type"]),
            # within each of those categories, reverse the 'as' order
            -first_index_of(type=title["type"], _as=title.get("as")),
            # put titles for portions last, within the type/as category
            title["is_for_portion"],
            # and within that, just sort alphabetically, case-insensitively (which is
            # what it appears THOMAS used to do)
            title["title"].lower(),
        )
    )

    return titles


# the most current title of a given type is the first one in the last 'as' subgroup
# of the titles for the whole bill (that is, if there's no title for the whole bill
# in the last 'as' subgroup, use the previous 'as' subgroup and so on) --- we think
# this logic matches THOMAS/Congress.gov.


def current_title_for(titles, title_type):
    current_title = None
    current_as = -1  # not None, cause for popular titles, None is a valid 'as'

    for title in titles:
        if title["type"] != title_type or title["is_for_portion"] == True:
            continue
        if title["as"] == current_as:
            continue
        # right type, new 'as', store first one
        current_title = title["title"]
        current_as = title["as"]

    return current_title


def actions_for(action_list, bill_id, title):
    # The bulk XML data has action history information from multiple sources. For
    # major actions, the Library of Congress (code 9) action item often duplicates
    # the information of a House/Senate action item. We have to skip one so that we
    # don't tag multiple history items with the same parsed action info, which
    # would imply the action (like a vote) ocurred multiple times. THOMAS appears
    # to have suppressed the Library of Congress action lines in certain cases
    # to avoid duplication - they were not in our older data files.
    #
    # Also, there are some ghost action items with totally empty text. Remove those.
    # TODO: When removed from upstream data, we can remove that check.
    closure = {
        "prev": None,
    }

    def keep_action(item, closure):
        if not item.get("text"):
            return False

        keep = True
        if closure["prev"]:
            if item["sourceSystem"].get("code") == "9":
                # Date must match previous action..
                # If both this and previous have a time, the times must match.
                # The text must approximately match. Sometimes the LOC text has a prefix
                #   and different whitespace. And they may drop references -- so we'll
                # use our action_for helper function to drop references from both
                # prior to the string comparison.
                if (
                    item["actionDate"] == closure["prev"]["actionDate"]
                    and (
                        item.get("actionTime") == closure["prev"].get("actionTime")
                        or not item.get("actionTime")
                        or not closure["prev"].get("actionTime")
                    )
                    and action_for(item)["text"]
                    .replace(" ", "")
                    .endswith(action_for(closure["prev"])["text"].replace(" ", ""))
                ):

                    keep = False
        closure["prev"] = item
        return keep

    action_list = [item for item in action_list if keep_action(item, closure)]

    # Turn the actions into dicts. The actions are in reverse-chronological
    # order in the bulk data XML. Process them in chronological order so that
    # our bill status logic sees the actions in the right order.

    def build_dict(item, closure):
        action_dict = action_for(item)

        extra_action_info, new_status = parse_bill_action(
            action_dict, closure["prev_status"], bill_id, title
        )

        # only change/reflect status change if there was one
        if new_status:
            action_dict["status"] = new_status
            closure["prev_status"] = new_status

        # add additional parsed fields
        if extra_action_info:
            action_dict.update(extra_action_info)

        return action_dict

    closure = {
        "prev_status": "INTRODUCED",
    }
    return [build_dict(action, closure) for action in reversed(action_list)]


# clean text, pull out the action type, any other associated metadata with an action
def action_for(item):
    # acted_at

    if not item.get("actionTime"):
        acted_at = item.get("actionDate", "")
    else:
        # Although we get the action date & time in an ISO-ish format (split
        # across two fields), and although we know it's in local time at the
        # U.S. Capitol (i.e. U.S. Eastern), we don't know the UTC offset which
        # is a part of how we used to serialize the time. So parse and then
        # use pytz (via format_datetime) to re-serialize.
        acted_at = utils.format_datetime(
            datetime.datetime.strptime(
                item.get("actionDate", "") + " " + item["actionTime"],
                "%Y-%m-%d %H:%M:%S",
            )
        )

    # text & references
    # (amendment actions don't always have text?)

    text = item["text"] if item.get("text") is not None else ""

    # strip out links
    text = re.sub(r"</?[Aa]( \S.*?)?>", "", text)

    # remove and extract references
    references = []
    match = re.search("\s*\(([^)]+)\)\s*$", text)
    if match:
        # remove the matched section
        text = text[0 : match.start()] + text[match.end() :]

        types = match.group(1)

        # fix use of comma or colon instead of a semi colon between reference types
        # have seen some accidental capitalization combined with accidental comma, thus the 'T'
        # e.g. "text of Title VII as reported in House: CR H3075-3077, Text omission from Title VII:" (hr5384-109)
        types = re.sub("[,:] ([a-zT])", r"; \1", types)
        # fix "CR:"
        types = re.sub("CR:", "CR", types)
        # fix a missing semicolon altogether between references
        # e.g. sres107-112, "consideration: CR S1877-1878 text as"
        types = re.sub("(\d+) +([a-z])", r"\1; \2", types)

        for reference in re.split("; ?", types):
            if ": " not in reference:
                type, reference = None, reference
            else:
                type, reference = reference.split(": ", 1)

            references.append({"type": type, "reference": reference})

    # extract committee IDs
    if item.get("committee"):
        # Data format through Dec. 13, 2019 had only one <committee/> (though node could be empty).
        committee_nodes = [item["committee"]]
    elif item.get("committees"):
        # Starting on Dec. 13, 2019, and with a slow rollout, multiple committees could be specified.
        # Thankfully our JSON output format allowed it already.
        committee_nodes = item["committees"].get("item", [])
    else:
        # <committee/> or <committees/>, whichever was present, was empty
        committee_nodes = []

    # form dict

    action_dict = {
        "acted_at": acted_at,
        "action_code": item.get("actionCode", ""),
        "committees": (
            [
                committee_item["systemCode"][0:-2].upper()
                for committee_item in committee_nodes
            ]
            if committee_nodes
            else None
        ),  # if empty, store None
        "references": references,
        "type": "action",  # replaced by parse_bill_action if a regex matches
        "text": text,
    }

    if not action_dict["committees"]:
        # remove if empty - not present in how we used to generate the file
        del action_dict["committees"]

    # sometimes there are links (one case is for bills passed by a rule in a resolution, the link will point to the resolution)
    if (item.get("links") or {}).get("link") is not None:
        action_dict["links"] = item["links"]["link"]

    return action_dict


def cosponsors_for(cosponsors_list):
    if cosponsors_list is None:
        return []

    cosponsors_list = cosponsors_list["item"]

    def build_dict(item):
        cosponsor_dict = sponsor_for(item)
        del cosponsor_dict["type"]  # always 'person'
        cosponsor_dict.update(
            {
                "sponsored_at": item["sponsorshipDate"],
                "withdrawn_at": item.get("sponsorshipWithdrawnDate"),
                "original_cosponsor": item["isOriginalCosponsor"] == "True",
            }
        )
        return cosponsor_dict

    cosponsors = [build_dict(cosponsor) for cosponsor in cosponsors_list]

    # TODO: Can remove. Sort like the old THOMAS order to make diffs easier.
    cosponsors.sort(key=lambda c: c["name"].lower())

    return cosponsors


def related_bills_for(related_bills_list):
    if related_bills_list is None:
        return []

    related_bills_list = related_bills_list["item"]

    def build_dict(item):

        return {
            "reason": item["relationshipDetails"]["item"][0]["type"]
            .replace("bill", "")
            .strip()
            .lower(),
            "bill_id": "{0}{1}-{2}".format(
                item["type"].replace(".", "").lower(), item["number"], item["congress"]
            ),
            "type": "bill",
            "identified_by": item["relationshipDetails"]["item"][0]["identifiedBy"],
        }

        # Are these THOMAS related bill relation texts gone from the bulk data?
        reasons = (
            ("Identical bill identified by (CRS|House|Senate)", "identical"),
            ("Companion bill", "identical"),
            (
                "Related bill (as )?identified by (CRS|the House Clerk's office|House committee|Senate)",
                "related",
            ),
            ("passed in (House|Senate) in lieu of .*", "supersedes"),
            ("Rule related to .* in (House|Senate)", "rule"),
            ("This bill has text inserted from .*", "includes"),
            ("Text from this bill was inserted in .*", "included-in"),
            ("Bill related to rule .* in House", "ruled-by"),
            ("This bill caused other related action on .*", "caused-action"),
            (
                "Other related action happened to this bill because of .*",
                "action-caused-by",
            ),
            ("Bill that causes .* to be laid on table in House", "caused-action"),
            ("Bill laid on table by virtue of .* passage in House", "action-caused-by"),
            ("Bill that caused the virtual passage of .* in House", "caused-action"),
            ("Bill passed by virtue of .* passage in House", "caused-action-by"),
            (
                "Bill on wich enrollment has been corrected by virtue of .* passage in House",
                "caused-action",
            ),
        )

    return [build_dict(related_bill) for related_bill in related_bills_list]


# get the public or private law number from any enacted action


def slip_law_from(actions):
    for action in actions:
        if action["type"] == "enacted":
            return {
                "law_type": action["law"],
                "congress": action["congress"],
                "number": action["number"],
            }


# find the latest status change in a set of processed actions


def latest_status(actions, introduced_at):
    status, status_date = "INTRODUCED", introduced_at
    for action in actions:
        if action.get("status", None):
            status = action["status"]
            status_date = action["acted_at"]
    return status, status_date


# look at the final set of processed actions and pull out the major historical events


def history_from_actions(actions):

    history = {}

    activation = activation_from(actions)
    if activation:
        history["active"] = True
        history["active_at"] = activation["acted_at"]
    else:
        history["active"] = False

    house_vote = None
    for action in actions:
        if (
            (action["type"] == "vote")
            and (action["where"] == "h")
            and (action["vote_type"] != "override")
        ):
            house_vote = action
    if house_vote:
        history["house_passage_result"] = house_vote["result"]
        history["house_passage_result_at"] = house_vote["acted_at"]

    senate_vote = None
    for action in actions:
        if (
            (action["type"] == "vote")
            and (action["where"] == "s")
            and (action["vote_type"] != "override")
        ):
            senate_vote = action
    if senate_vote:
        history["senate_passage_result"] = senate_vote["result"]
        history["senate_passage_result_at"] = senate_vote["acted_at"]

    senate_vote = None
    for action in actions:
        if (
            (action["type"] == "vote-aux")
            and (action["vote_type"] == "cloture")
            and (action["where"] == "s")
            and (action["vote_type"] != "override")
        ):
            senate_vote = action
    if senate_vote:
        history["senate_cloture_result"] = senate_vote["result"]
        history["senate_cloture_result_at"] = senate_vote["acted_at"]

    vetoed = None
    for action in actions:
        if action["type"] == "vetoed":
            vetoed = action
    if vetoed:
        history["vetoed"] = True
        history["vetoed_at"] = vetoed["acted_at"]
    else:
        history["vetoed"] = False

    house_override_vote = None
    for action in actions:
        if (
            (action["type"] == "vote")
            and (action["where"] == "h")
            and (action["vote_type"] == "override")
        ):
            house_override_vote = action
    if house_override_vote:
        history["house_override_result"] = house_override_vote["result"]
        history["house_override_result_at"] = house_override_vote["acted_at"]

    senate_override_vote = None
    for action in actions:
        if (
            (action["type"] == "vote")
            and (action["where"] == "s")
            and (action["vote_type"] == "override")
        ):
            senate_override_vote = action
    if senate_override_vote:
        history["senate_override_result"] = senate_override_vote["result"]
        history["senate_override_result_at"] = senate_override_vote["acted_at"]

    enacted = None
    for action in actions:
        if action["type"] == "enacted":
            enacted = action
    if enacted:
        history["enacted"] = True
        history["enacted_at"] = action["acted_at"]
    else:
        history["enacted"] = False

    topresident = None
    for action in actions:
        if action["type"] == "topresident":
            topresident = action
    if topresident and (not history["vetoed"]) and (not history["enacted"]):
        history["awaiting_signature"] = True
        history["awaiting_signature_since"] = action["acted_at"]
    else:
        history["awaiting_signature"] = False

    return history


# find the first action beyond the standard actions every bill gets.
# - if the bill's first action is "referral" then the first action not those
#     most common
#     e.g. hr3590-111 (active), s1-113 (inactive)
# - if the bill's first action is "action", then the next action, if one is present
#     resolutions
#     e.g. sres5-113 (active), sres4-113 (inactive)
# - if the bill's first action is anything else (e.g. "vote"), then that first action
#     bills that skip committee
#     e.g. s227-113 (active)
def activation_from(actions):
    # there's NOT always at least one :(
    # as of 2013-06-10, hr2272-113 has no actions at all
    if len(actions) == 0:
        return None

    first = actions[0]

    if first["type"] in ["referral", "calendar", "action"]:
        for action in actions[1:]:
            if (
                (action["type"] != "referral")
                and (action["type"] != "calendar")
                and ("Sponsor introductory remarks" not in action["text"])
            ):
                return action
        return None
    else:
        return first


def parse_bill_action(action_dict, prev_status, bill_id, title):
    """Parse a THOMAS bill action line. Returns attributes to be set in the XML file on the action line."""

    bill_type, number, congress = utils.split_bill_id(bill_id)

    line = action_dict["text"]

    status = None
    action = {"type": "action"}

    # If a line starts with an amendment number, this action is on the amendment and cannot
    # be parsed yet.
    m = re.search(r"^(H|S)\.Amdt\.(\d+)", line, re.I)
    if m != None:
        # Process actions specific to amendments separately.
        return None, None

    # Otherwise, parse the action line for key actions.

    # VOTES

    # A House Vote.
    line = re.sub(", the Passed", ", Passed", line)  # 106 h4733 and others
    m = re.search(
        "("
        + "|".join(
            [
                "On passage",
                "Passed House",
                "Two-thirds of the Members present having voted in the affirmative the bill is passed,?",
                "On motion to suspend the rules and pass the (?:bill|resolution)",
                "On agreeing to the (?:resolution|conference report)",
                "On motion to suspend the rules and agree to the (?:resolution|conference report)",
                "House Agreed to Senate Amendments.*?",
                "On motion (?:that )?the House (?:suspend the rules and )?(?:agree(?: with an amendment)? to|concur in) the Senate amendments?(?: to the House amendments?| to the Senate amendments?)*",
            ]
        )
        + ")"
        + "(, the objections of the President to the contrary notwithstanding.?)?"
        + "(, as amended| \(Amended\))?"
        + "\.? (Passed|Failed|Agreed to|Rejected)?"  # hr1625-115 has a stray period here
        + " ?(by voice vote|without objection|by (the Yeas and Nays?|Yea-Nay Vote|recorded vote)"
        + "(:? \(2/3 required\))?: (\d+ ?- ?\d+(, \d+ Present)? [ \)]*)?\((Roll no\.|Record Vote No:) \d+\))",
        line,
        re.I,
    )
    if m != None:
        motion, is_override, as_amended, pass_fail, how = (
            m.group(1),
            m.group(2),
            m.group(3),
            m.group(4),
            m.group(5),
        )

        # print(line)
        # print(m.groups())

        if re.search(r"Passed House|House Agreed to", motion, re.I):
            pass_fail = "pass"
        elif re.search("(ayes|yeas) had prevailed", line, re.I):
            pass_fail = "pass"
        elif re.search(r"Pass|Agreed", pass_fail, re.I):
            pass_fail = "pass"
        else:
            pass_fail = "fail"

        if "Two-thirds of the Members present" in motion:
            is_override = True

        if is_override:
            vote_type = "override"
        elif re.search(
            r"(agree (with an amendment )?to|concur in) the Senate amendment",
            line,
            re.I,
        ):
            vote_type = "pingpong"
        elif re.search("conference report", line, re.I):
            vote_type = "conference"
        elif bill_type[0] == "h":
            vote_type = "vote"
        else:
            vote_type = "vote2"

        roll = None
        m = re.search(r"\((Roll no\.|Record Vote No:) (\d+)\)", how, re.I)
        if m != None:
            how = "roll"  # normalize the ugly how
            roll = m.group(2)

        suspension = None
        if roll and "On motion to suspend the rules" in motion:
            suspension = True

        # alternate form of as amended, e.g. hr3979-113
        if "the House agree with an amendment" in motion:
            as_amended = True

        action["type"] = "vote"
        action["vote_type"] = vote_type
        action["how"] = how
        action["where"] = "h"
        action["result"] = pass_fail
        if roll:
            action["roll"] = roll
        action["suspension"] = suspension

        # correct upstream data error
        if bill_id == "s2012-114" and "Roll no. 250" in line:
            as_amended = True
        if bill_id == "s2943-114" and "On passage Passed without objection" in line:
            as_amended = True

        # get the new status of the bill after this vote
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            "h",
            bill_type,
            suspension,
            as_amended,
            title,
            prev_status,
        )
        if new_status:
            status = new_status

    # Passed House, not necessarily by an actual vote (think "deem")
    m = re.search(
        r"Passed House pursuant to|House agreed to Senate amendment (with amendment )?pursuant to|Pursuant to the provisions of [HSCONJRES\. ]+ \d+, [HSCONJRES\. ]+ \d+ is considered passed House",
        line,
        re.I,
    )
    if m != None:
        vote_type = "vote" if (bill_type[0] == "h") else "vote2"
        if "agreed to Senate amendment" in line:
            vote_type = "pingpong"
        pass_fail = "pass"
        as_amended = ("with amendment" in line) or ("as amended" in line)

        action["type"] = "vote"
        action["vote_type"] = vote_type
        action["how"] = "by special rule"
        action["where"] = "h"
        action["result"] = pass_fail

        # It's always pursuant to another bill, and a bill number is given in the action line, which we parse out
        # into the bill_ids field of the action. It's also represented
        # structurally in the links->link elements of the original XML which we just put in "links".

        # get the new status of the bill after this vote
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            "h",
            bill_type,
            False,
            as_amended,
            title,
            prev_status,
        )

        if new_status:
            status = new_status

    m = re.search(
        r"Pursuant to .* the following bills passed under suspension of the rules: (.*)\.$",
        line,
        re.I,
    )
    if m:
        # The list should certainly include this bill, but was it passed "as amended"?
        as_amended = None
        bill_list = m.group(1)
        bill_list = bill_list.replace(
            "and the following resolution was agreed to under suspension of the rules: ",
            "",
        )
        bill_list = bill_list.replace(
            "and the following resolutions were agreed to under suspension of the rules: ",
            "",
        )
        bill_list = bill_list.replace("and ", "")
        bill_list = re.split(r"\s*(?:;|,(?! as amended))\s*", bill_list)
        for bill_item in bill_list:
            bill_item = bill_item.lower().replace(".", "").replace(" ", "").split(",")
            if bill_item[0] == (bill_type + number):
                as_amended = len(bill_item) > 1
        if as_amended is not None:  # found the bill in the list?
            vote_type = "vote" if (bill_type[0] == "h") else "vote2"
            pass_fail = "pass"
            action["type"] = "vote"
            action["vote_type"] = vote_type
            action["how"] = "by special rule"
            action["where"] = "h"
            action["result"] = pass_fail
            new_status = new_status_after_vote(
                vote_type,
                pass_fail == "pass",
                "h",
                bill_type,
                False,
                as_amended,
                title,
                prev_status,
            )
            if new_status:
                status = new_status

    # House motions to table adversely dispose of a pending matter, if agreed to. An agreed-to "motion to table the measure",
    # which is very infrequent, kills the legislation. If not agreed to, nothing changes. So this regex only captures
    # agreed-to motions to table.
    m = re.search(
        "On motion to table the measure Agreed to"
        + " ?(by voice vote|without objection|by (the Yeas and Nays|Yea-Nay Vote|recorded vote)"
        + ": (\d+ - \d+(, \d+ Present)? [ \)]*)?\((Roll no\.|Record Vote No:) \d+\))",
        line,
        re.I,
    )
    if m != None:
        how = m.group(1)
        pass_fail = "fail"

        # In order to classify this as resulting in the same thing as regular failed vote on passage, new_status_after_vote
        # needs to know if this was a vote in the originating chamber or not.
        if prev_status in ("INTRODUCED", "REPORTED") or bill_id.startswith("hres"):
            vote_type = "vote"
        elif False:
            vote_type = "vote2"
        else:
            raise Exception(
                "Need to classify %s as being in the originating chamber or not."
                % prev_status
            )

        roll = None
        m = re.search(r"\((Roll no\.|Record Vote No:) (\d+)\)", how, re.I)
        if m != None:
            how = "roll"  # normalize the ugly how
            roll = m.group(2)

        action["type"] = "vote"
        action["vote_type"] = vote_type
        action["how"] = how
        action["where"] = "h"
        action["result"] = pass_fail
        if roll:
            action["roll"] = roll

        # get the new status of the bill after this vote
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            "h",
            bill_type,
            False,
            False,
            title,
            prev_status,
        )
        if new_status:
            status = new_status

    # A Senate Vote
    # (There are some annoying weird cases of double spaces which are taken care of
    # at the end.)
    m = re.search(
        "("
        + "|".join(
            [
                "Passed Senate",
                "Failed of passage in Senate",
                "Disagreed to in Senate",
                "Resolution agreed to in Senate",
                "Senate (?:agreed to|concurred in) (?:the )?(?:conference report|House amendment(?: to the Senate amendments?| to the House amendments?)*)",
                "Senate receded from its amendment and concurred",  # hr1-115
                r"Cloture \S*\s?on the motion to proceed .*?not invoked in Senate",
                r"Cloture(?: motion)? on the motion to proceed to the (?:bill|measure) invoked in Senate",
                "Cloture invoked in Senate",
                "Cloture on (?:the motion to (?:proceed to |concur in )(?:the House amendment (?:to the Senate amendment )?to )?)(?:the bill|H.R. .*) (?:not )?invoked in Senate",
                "(?:Introduced|Received|Submitted) in the Senate, (?:read twice, |considered, |read the third time, )+and (?:passed|agreed to)",
            ]
        )
        + ")"
        + "(,?.*,?) "
        + "(without objection|by Unanimous Consent|by Voice Vote|(?:by )?Yea-Nay( Vote)?\. \d+\s*-\s*\d+\. Record Vote (No|Number): \d+)",
        line.replace("  ", " "),
        re.I,
    )
    if m != None:
        motion, extra, how = m.group(1), m.group(2), m.group(3)
        roll = None

        # put disagreed check first, cause "agreed" is contained inside it
        if re.search("disagreed|not invoked", motion, re.I):
            pass_fail = "fail"
        elif re.search("passed|agreed|concurred|invoked", motion, re.I):
            pass_fail = "pass"
        else:
            pass_fail = "fail"

        voteaction_type = "vote"
        if re.search("over veto", extra, re.I):
            vote_type = "override"
        elif re.search("conference report", motion, re.I):
            vote_type = "conference"
        elif re.search("cloture", motion, re.I):
            vote_type = "cloture"
            voteaction_type = "vote-aux"  # because it is not a vote on passage
        elif re.search(
            "Senate agreed to (the )?House amendment|Senate concurred in (the )?House amendment",
            motion,
            re.I,
        ):
            vote_type = "pingpong"
        elif bill_type[0] == "s":
            vote_type = "vote"
        else:
            vote_type = "vote2"

        m = re.search(r"Record Vote (No|Number): (\d+)", how, re.I)
        if m != None:
            roll = m.group(2)
            how = "roll"

        as_amended = False
        if re.search(r"with amendments|with an amendment", extra, re.I):
            as_amended = True

        action["type"] = voteaction_type
        action["vote_type"] = vote_type
        action["how"] = how
        action["result"] = pass_fail
        action["where"] = "s"
        if roll:
            action["roll"] = roll

        # get the new status of the bill after this vote
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            "s",
            bill_type,
            False,
            as_amended,
            title,
            prev_status,
        )

        if new_status:
            status = new_status

    # OLD-STYLE VOTES (93rd Congress-ish)

    m = re.search(
        r"Measure passed (House|Senate)(, amended(?: \(.*?\)|, with an amendment to the title)?)?(?:,? in lieu[^,]*)?(?:, roll call #(\d+) \(\d+-\d+\))?",
        line,
        re.I,
    )
    if m != None:
        chamber = m.group(1)[0].lower()  # 'h' or 's'
        as_amended = m.group(2)
        roll_num = m.group(3)
        # GovTrack legacy scraper missed these: if chamber == 's' and (as_amended or roll_num or "lieu" in line): return action, status
        pass_fail = "pass"
        vote_type = "vote" if bill_type[0] == chamber else "vote2"
        action["type"] = "vote"
        action["vote_type"] = vote_type
        action["how"] = "(method not recorded)" if not roll_num else "roll"
        if roll_num:
            action["roll"] = roll_num
        action["result"] = pass_fail
        action["where"] = chamber
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            chamber,
            bill_type,
            False,
            as_amended,
            title,
            prev_status,
        )
        if new_status:
            status = new_status

    m = re.search(
        r"(House|Senate) agreed to (?:House|Senate) amendments?( with an amendment)?( under Suspension of the Rules)?(?:, roll call #(\d+) \(\d+-\d+\))?\.",
        line,
        re.I,
    )
    if m != None:
        chamber = m.group(1)[0].lower()  # 'h' or 's'
        as_amended = m.group(2)
        suspension = m.group(3)
        roll_num = m.group(4)
        # GovTrack legacy scraper missed these: if (chamber == 'h' and not roll_num) or (chamber == 's' and rull_num): return action, status # REMOVE ME
        pass_fail = "pass"
        vote_type = "pingpong"
        action["type"] = "vote"
        action["vote_type"] = vote_type
        action["how"] = "(method not recorded)" if not roll_num else "roll"
        if roll_num:
            action["roll"] = roll_num
        action["result"] = pass_fail
        action["where"] = chamber
        action["suspension"] = suspension != None
        new_status = new_status_after_vote(
            vote_type,
            pass_fail == "pass",
            chamber,
            bill_type,
            False,
            as_amended,
            title,
            prev_status,
        )
        if new_status:
            status = new_status

    # Useless. But GovTrack has had it.
    m = re.search(
        r"Placed on (the )?([\w ]+) Calendar( under ([\w ]+))?[,\.] Calendar No\. (\d+)\.",
        line,
        re.I,
    )
    if m != None:
        action["type"] = "calendar"
        action["calendar"] = m.group(2)
        action["under"] = m.group(4)
        action["number"] = m.group(5)

    # COMMITTEE ACTIONS

    # Ordered Reported (because GovTrack did this, but maybe should be changed to not combine with actual reported bills)
    m = re.search(
        r"Ordered to be Reported|Committee Agreed to Seek Consideration Under Suspension of the Rules",
        line,
        re.I,
    )
    if m != None:
        action["type"] = "ordered-reported"
        if prev_status in ("INTRODUCED", "REFERRED"):
            status = "REPORTED"

    # reported
    m = re.search(
        r"Committee on (.*)\. (Original measure )?[Rr]eported (to Senate )?by",
        line,
        re.I,
    )
    if m != None:
        action["type"] = "reported"
        action["committee"] = m.group(1)
        if prev_status in ("INTRODUCED", "REFERRED"):
            status = "REPORTED"
    m = re.search(
        r"Reported to Senate from the (.*?)( \(without written report\))?\.", line, re.I
    )
    if m != None:  # 93rd Congress
        action["type"] = "reported"
        action["committee"] = m.group(1)
        if prev_status in ("INTRODUCED", "REFERRED"):
            status = "REPORTED"

    # hearings held by a committee
    m = re.search(r"(Committee on .*?)\. Hearings held", line, re.I)
    if m != None:
        action["committee"] = m.group(1)
        action["type"] = "hearings"

    m = re.search(r"Committee on (.*)\. Discharged (by Unanimous Consent)?", line, re.I)
    if m != None:
        action["committee"] = m.group(1)
        action["type"] = "discharged"
        if prev_status in ("INTRODUCED", "REFERRED"):
            status = "REPORTED"

    m = re.search("Cleared for White House|Presented to President", line, re.I)
    if m != None:
        action["type"] = "topresident"

    m = re.search("Signed by President", line, re.I)
    if m != None:
        action["type"] = "signed"
        status = "ENACTED:SIGNED"

    m = re.search("Pocket Vetoed by President", line, re.I)
    if m != None:
        action["type"] = "vetoed"
        action["pocket"] = "1"
        status = "VETOED:POCKET"

    # need to put this in an else, or this regex will match the pocket veto and override it
    else:
        m = re.search("Vetoed by President", line, re.I)
        if m != None:
            action["type"] = "vetoed"
            status = "PROV_KILL:VETO"

    m = re.search("Sent to Archivist of the United States unsigned", line, re.I)
    if m != None:
        status = "ENACTED:TENDAYRULE"

    m = re.search("^(?:Became )?(Public|Private) Law(?: No:)? ([\d\-]+)\.", line, re.I)
    if m != None:
        action["law"] = m.group(1).lower()
        pieces = m.group(2).split("-")
        action["congress"] = pieces[0]
        action["number"] = pieces[1]
        action["type"] = "enacted"
        if prev_status in (
            "ENACTED:SIGNED",
            "ENACTED:VETO_OVERRIDE",
            "ENACTED:TENDAYRULE",
        ):
            pass  # this is a final administrative step
        elif prev_status == "PROV_KILL:VETO" or prev_status.startswith("VETOED:"):
            # somehow missed the override steps
            status = "ENACTED:VETO_OVERRIDE"
        elif bill_id in (
            "s2641-93",
            "hr1589-94",
            "s2527-100",
            "hr1677-101",
            "hr2978-101",
            "hr2126-104",
            "s1322-104",
        ):
            status = "ENACTED:TENDAYRULE"
        else:
            raise Exception(
                "Missing Signed by President action? If this is a case of the 10-day rule, hard code the bill id %s here."
                % bill_id
            )

    # Check for referral type
    m = re.search(
        r"Referred to (?:the )?(House|Senate)?\s?(?:Committee|Subcommittee)?",
        line,
        re.I,
    )
    if m != None:
        action["type"] = "referral"
        if prev_status == "INTRODUCED":
            status = "REFERRED"

    # sweep the action line for bill IDs of related bills
    bill_ids = utils.extract_bills(line, congress)
    bill_ids = [b for b in bill_ids if b != bill_id]
    if bill_ids and (len(bill_ids) > 0):
        action["bill_ids"] = bill_ids

    return action, status


def new_status_after_vote(
    vote_type, passed, chamber, bill_type, suspension, amended, title, prev_status
):
    if vote_type == "vote":  # vote in originating chamber
        if passed:
            if bill_type in ("hres", "sres"):
                return "PASSED:SIMPLERES"  # end of life for a simple resolution
            if chamber == "h":
                return "PASS_OVER:HOUSE"  # passed by originating chamber, now in second chamber
            else:
                return "PASS_OVER:SENATE"  # passed by originating chamber, now in second chamber
        if suspension:
            return "PROV_KILL:SUSPENSIONFAILED"  # provisionally killed by failure to pass under suspension of the rules
        if chamber == "h":
            return "FAIL:ORIGINATING:HOUSE"  # outright failure
        else:
            return "FAIL:ORIGINATING:SENATE"  # outright failure
    if vote_type in (
        "vote2",
        "pingpong",
    ):  # vote in second chamber or subsequent pingpong votes
        if passed:
            if amended:
                # mesure is passed but not in identical form
                if chamber == "h":
                    return "PASS_BACK:HOUSE"  # passed both chambers, but House sends it back to Senate
                else:
                    return "PASS_BACK:SENATE"  # passed both chambers, but Senate sends it back to House
            else:
                # bills and joint resolutions not constitutional amendments, not amended from Senate version
                if bill_type in ("hjres", "sjres") and title.startswith(
                    "Proposing an amendment to the Constitution of the United States"
                ):
                    return "PASSED:CONSTAMEND"  # joint resolution that looks like an amendment to the constitution
                if bill_type in ("hconres", "sconres"):
                    return (
                        "PASSED:CONCURRENTRES"  # end of life for concurrent resolutions
                    )
                return "PASSED:BILL"  # passed by second chamber, now on to president
        if vote_type == "pingpong":
            # chamber failed to accept the other chamber's changes, but it can vote again
            return "PROV_KILL:PINGPONGFAIL"
        if suspension:
            return "PROV_KILL:SUSPENSIONFAILED"  # provisionally killed by failure to pass under suspension of the rules
        if chamber == "h":
            return "FAIL:SECOND:HOUSE"  # outright failure
        else:
            return "FAIL:SECOND:SENATE"  # outright failure
    if vote_type == "cloture":
        if not passed:
            return "PROV_KILL:CLOTUREFAILED"
        else:
            return None
    if vote_type == "override":
        if not passed:
            if bill_type[0] == chamber:
                if chamber == "h":
                    return "VETOED:OVERRIDE_FAIL_ORIGINATING:HOUSE"
                else:
                    return "VETOED:OVERRIDE_FAIL_ORIGINATING:SENATE"
            else:
                if chamber == "h":
                    return "VETOED:OVERRIDE_FAIL_SECOND:HOUSE"
                else:
                    return "VETOED:OVERRIDE_FAIL_SECOND:SENATE"
        else:
            if bill_type[0] == chamber:
                if chamber == "h":
                    return "VETOED:OVERRIDE_PASS_OVER:HOUSE"
                else:
                    return "VETOED:OVERRIDE_PASS_OVER:SENATE"
            else:
                # The override passed both chambers -- the veto is overridden.
                return "ENACTED:VETO_OVERRIDE"
    if vote_type == "conference":
        # This is tricky to integrate into status because we have to wait for both
        # chambers to pass the conference report.
        if passed:
            if prev_status.startswith("CONFERENCE:PASSED:"):
                if bill_type in ("hjres", "sjres") and title.startswith(
                    "Proposing an amendment to the Constitution of the United States"
                ):
                    return "PASSED:CONSTAMEND"  # joint resolution that looks like an amendment to the constitution
                if bill_type in ("hconres", "sconres"):
                    return (
                        "PASSED:CONCURRENTRES"  # end of life for concurrent resolutions
                    )
                return "PASSED:BILL"
            else:
                if chamber == "h":
                    return "CONFERENCE:PASSED:HOUSE"
                else:
                    return "CONFERENCE:PASSED:SENATE"

    return None


def amendments_for(amendment_list):
    if amendment_list is None:
        return []

    amendment_list = amendment_list["amendment"]

    def build_dict(item):
        # Malformed XML containing duplicate elements causes attributes to parse as a list
        for attr in ["type", "number", "congress"]:
            if type(item[attr]) is list:
                item[attr] = item[attr][0]
        return {
            "amendment_id": "{0}{1}-{2}".format(
                item["type"].lower(), item["number"], item["congress"]
            ),
            "amendment_type": item["type"].lower(),
            "chamber": item["type"][0].lower(),
            "number": item["number"],
        }

    return [build_dict(amendment) for amendment in amendment_list]


def committee_reports_for(committeeReports):
    ret = []
    for report in (committeeReports or {}).get("committeeReport", []):
        ret.append(report["citation"])
    return ret
